#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import copy
import json
from unittest import mock

from cinderclient import exceptions as cinder_exp
from novaclient import exceptions as nova_exp
from oslo_config import cfg

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import cinder
from heat.engine.clients.os import glance
from heat.engine.clients.os import nova
from heat.engine.resources.openstack.cinder import volume as c_vol
from heat.engine.resources import scheduler_hints as sh
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.objects import resource_data as resource_data_object
from heat.tests.openstack.cinder import test_volume_utils as vt_base
from heat.tests.openstack.nova import fakes as fakes_nova
from heat.tests import utils

cinder_volume_template = '''
heat_template_version: 2013-05-23
description: Cinder volumes and attachments.
resources:
  volume:
    type: OS::Cinder::Volume
    properties:
      availability_zone: nova
      size: 1
      name: test_name
      description: test_description
      metadata:
        key: value
  volume2:
    type: OS::Cinder::Volume
    properties:
      availability_zone: nova
      size: 2
  volume3:
    type: OS::Cinder::Volume
    properties:
      availability_zone: nova
      size: 1
      name: test_name
      scheduler_hints: {"hint1": "good_advice"}
  volume4:
    type: OS::Cinder::Volume
    properties:
      availability_zone: nova
      size: 1
      name: test_name
  attachment:
    type: OS::Cinder::VolumeAttachment
    properties:
      instance_uuid: WikiDatabase
      volume_id: { get_resource: volume }
      mountpoint: /dev/vdc
'''

single_cinder_volume_template = '''
heat_template_version: 2013-05-23
description: Cinder volume
resources:
  volume:
    type: OS::Cinder::Volume
    properties:
      size: 1
      name: test_name
      description: test_description
'''


class CinderVolumeTest(vt_base.VolumeTestCase):

    def setUp(self):
        super(CinderVolumeTest, self).setUp()
        self.t = template_format.parse(cinder_volume_template)
        self.use_cinder = True

    def create_volume(self, t, stack, resource_name,
                      mock_check_extend_ready=True):
        rsrc = super(CinderVolumeTest, self).create_volume(
            t, stack, resource_name, validate=False)
        if mock_check_extend_ready:
            rsrc._ready_to_extend_volume = mock.Mock(return_value=True)
        return rsrc

    def _mock_create_volume(self, fv, stack_name, size=1,
                            final_status='available', extra_get_mocks=[],
                            extra_create_mocks=[]):
        result = [fv]
        for m in extra_create_mocks:
            result.append(m)
        self.cinder_fc.volumes.create.side_effect = result
        fv_ready = vt_base.FakeVolume(final_status, id=fv.id)
        result = [fv, fv_ready]
        for m in extra_get_mocks:
            result.append(m)
        self.cinder_fc.volumes.get.side_effect = result
        return fv_ready

    def test_cinder_volume_size_constraint(self):
        self.t['resources']['volume']['properties']['size'] = 0
        stack = utils.parse_stack(self.t)
        error = self.assertRaises(exception.StackValidationFailed,
                                  self.create_volume,
                                  self.t, stack, 'volume')
        self.assertEqual(
            "Property error: resources.volume.properties.size: "
            "0 is out of range (min: 1, max: None)", str(error))

    def test_cinder_create(self):
        fv = vt_base.FakeVolume('creating')
        self.stack_name = 'test_cvolume_stack'

        self.stub_SnapshotConstraint_validate()
        self.stub_VolumeConstraint_validate()
        self.stub_VolumeTypeConstraint_validate()
        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]

        self.t['resources']['volume']['properties'].update({
            'volume_type': 'lvm',
        })
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        self.create_volume(self.t, stack, 'volume')

        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, availability_zone='nova',
            description='test_description',
            name='test_name',
            metadata={'key': 'value'},
            volume_type='lvm')
        self.assertEqual(2, self.cinder_fc.volumes.get.call_count)

    def test_cinder_create_from_image(self):
        fv = vt_base.FakeVolume('downloading')
        self.stack_name = 'test_cvolume_create_from_img_stack'
        image_id = '46988116-6703-4623-9dbc-2bc6d284021b'
        self.patchobject(glance.GlanceClientPlugin,
                         'find_image_by_name_or_id', return_value=image_id)

        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]

        self.t['resources']['volume']['properties'] = {
            'size': '1',
            'name': 'ImageVolume',
            'description': 'ImageVolumeDescription',
            'availability_zone': 'nova',
            'image': image_id,
        }
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        self.create_volume(self.t, stack, 'volume')

        glance.GlanceClientPlugin.find_image_by_name_or_id.assert_called_with(
            image_id)
        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, availability_zone='nova',
            description='ImageVolumeDescription',
            name='ImageVolume',
            imageRef=image_id,
            metadata={})
        self.assertEqual(2, self.cinder_fc.volumes.get.call_count)

    def test_cinder_create_with_read_only(self):
        fv = vt_base.FakeVolume('with_read_only_access_mode')
        self.stack_name = 'test_create_with_read_only'

        self.cinder_fc.volumes.create.return_value = fv

        update_readonly_mock = self.patchobject(self.cinder_fc.volumes,
                                                'update_readonly_flag',
                                                return_value=None)
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.return_value = fv_ready

        self.t['resources']['volume']['properties'] = {
            'size': '1',
            'name': 'ImageVolume',
            'description': 'ImageVolumeDescription',
            'availability_zone': 'nova',
            'read_only': False,
        }
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        self.create_volume(self.t, stack, 'volume')

        update_readonly_mock.assert_called_once_with(fv.id, False)

        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, availability_zone='nova',
            description='ImageVolumeDescription',
            name='ImageVolume',
            metadata={})
        self.cinder_fc.volumes.get.assert_called_once_with(fv.id)

    def test_cinder_default(self):
        fv = vt_base.FakeVolume('creating')
        self.stack_name = 'test_cvolume_default_stack'

        vol_name = utils.PhysName(self.stack_name, 'volume')
        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]

        self.t['resources']['volume']['properties'] = {
            'size': '1',
            'availability_zone': 'nova',
        }
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        self.create_volume(self.t, stack, 'volume')

        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, availability_zone='nova',
            description=None,
            name=vol_name,
            metadata={}
        )
        self.assertEqual(2, self.cinder_fc.volumes.get.call_count)

    def test_cinder_fn_getatt(self):
        self.stack_name = 'test_cvolume_fngetatt_stack'

        fv = vt_base.FakeVolume(
            'available', availability_zone='zone1',
            size=1, snapshot_id='snap-123', name='name',
            description='desc', volume_type='lvm',
            metadata={'key': 'value'}, source_volid=None,
            bootable=False, created_at='2013-02-25T02:40:21.000000',
            encrypted=False, attachments=[])

        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[fv for i in range(20)])

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        rsrc = self.create_volume(self.t, stack, 'volume')

        self.assertEqual('zone1', rsrc.FnGetAtt('availability_zone'))
        self.assertEqual('1', rsrc.FnGetAtt('size'))
        self.assertEqual('snap-123', rsrc.FnGetAtt('snapshot_id'))
        self.assertEqual('name', rsrc.FnGetAtt('display_name'))
        self.assertEqual('desc', rsrc.FnGetAtt('display_description'))
        self.assertEqual('lvm', rsrc.FnGetAtt('volume_type'))
        self.assertEqual(json.dumps({'key': 'value'}),
                         rsrc.FnGetAtt('metadata'))
        self.assertEqual({'key': 'value'},
                         rsrc.FnGetAtt('metadata_values'))
        self.assertEqual('None', rsrc.FnGetAtt('source_volid'))
        self.assertEqual('available', rsrc.FnGetAtt('status'))
        self.assertEqual('2013-02-25T02:40:21.000000',
                         rsrc.FnGetAtt('created_at'))
        self.assertEqual('False', rsrc.FnGetAtt('bootable'))
        self.assertEqual('False', rsrc.FnGetAtt('encrypted'))
        self.assertEqual('[]', rsrc.FnGetAtt('attachments'))
        self.assertEqual([], rsrc.FnGetAtt('attachments_list'))
        error = self.assertRaises(exception.InvalidTemplateAttribute,
                                  rsrc.FnGetAtt, 'unknown')
        self.assertEqual(
            'The Referenced Attribute (volume unknown) is incorrect.',
            str(error))

        self.cinder_fc.volumes.get.assert_called_with('vol-123')

    def test_cinder_attachment(self):
        self.stack_name = 'test_cvolume_attach_stack'
        fva = vt_base.FakeVolume('in-use')

        m_v = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'))
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     m_v, fva,
                                     vt_base.FakeVolume('available')])
        self.stub_VolumeConstraint_validate()

        # delete script
        self.fc.volumes.get_server_volume.side_effect = [
            fva, fva, fakes_nova.fake_exception()]

        self.fc.volumes.delete_server_volume.return_value = None

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        self.create_volume(self.t, stack, 'volume')
        rsrc = self.create_attachment(self.t, stack, 'attachment')
        scheduler.TaskRunner(rsrc.delete)()

        self.fc.volumes.get_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')
        self.fc.volumes.delete_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')

    def test_cinder_attachment_no_mountpoint(self):
        self.stack_name = 'test_cvolume_attach_stack'

        fva = vt_base.FakeVolume('in-use')
        m_v = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'))
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     m_v, fva,
                                     vt_base.FakeVolume('available')])
        self.stub_VolumeConstraint_validate()

        self.t['resources']['attachment']['properties']['mountpoint'] = ''
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        self.create_volume(self.t, stack, 'volume')
        rsrc = self.create_attachment(self.t, stack, 'attachment')

        # delete script
        self.fc.volumes.get_server_volume.side_effect = [
            fva, fva, fakes_nova.fake_exception()]
        self.fc.volumes.delete_server_volume.return_value = None

        scheduler.TaskRunner(rsrc.delete)()

        self.fc.volumes.get_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')
        self.fc.volumes.delete_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')

    def test_cinder_volume_shrink_fails(self):
        self.stack_name = 'test_cvolume_shrink_fail_stack'

        fv = vt_base.FakeVolume('available', size=2)
        # create script
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name, size=2, extra_get_mocks=[fv])

        self.t['resources']['volume']['properties']['size'] = 2
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume')

        props = copy.deepcopy(rsrc.properties.data)
        props['size'] = 1
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)

        update_task = scheduler.TaskRunner(rsrc.update, after)
        ex = self.assertRaises(exception.ResourceFailure, update_task)
        self.assertEqual('NotSupported: resources.volume: '
                         'Shrinking volume is not supported.',
                         str(ex))

        self.assertEqual((rsrc.UPDATE, rsrc.FAILED), rsrc.state)

    def test_cinder_volume_extend_detached(self):
        self.stack_name = 'test_cvolume_extend_det_stack'
        fv = vt_base.FakeVolume('available',
                                size=1, attachments=[])

        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     fv, vt_base.FakeVolume('extending'),
                                     vt_base.FakeVolume('extending'),
                                     vt_base.FakeVolume('available')])

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume')

        props = copy.deepcopy(rsrc.properties.data)
        props['size'] = 2
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)

        update_task = scheduler.TaskRunner(rsrc.update, after)
        self.assertIsNone(update_task())

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.cinder_fc.volumes.extend.assert_called_once_with(fv.id, 2)

    def test_cinder_volume_extend_fails_to_start(self):
        self.stack_name = 'test_cvolume_extend_fail_start_stack'

        fv = vt_base.FakeVolume('available',
                                size=1, attachments=[])
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[fv])
        self.cinder_fc.volumes.extend.side_effect = cinder_exp.OverLimit(413)

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume')

        props = copy.deepcopy(rsrc.properties.data)
        props['size'] = 2
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)

        update_task = scheduler.TaskRunner(rsrc.update, after)
        ex = self.assertRaises(exception.ResourceFailure, update_task)
        self.assertIn('Over limit', str(ex))

        self.assertEqual((rsrc.UPDATE, rsrc.FAILED), rsrc.state)
        self.cinder_fc.volumes.extend.assert_called_once_with(fv.id, 2)

    def test_cinder_volume_extend_fails_to_complete(self):
        self.stack_name = 'test_cvolume_extend_fail_compl_stack'

        fv = vt_base.FakeVolume('available',
                                size=1, attachments=[])
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     fv, vt_base.FakeVolume('extending'),
                                     vt_base.FakeVolume('extending'),
                                     vt_base.FakeVolume('error_extending')])

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume')

        props = copy.deepcopy(rsrc.properties.data)
        props['size'] = 2
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)

        update_task = scheduler.TaskRunner(rsrc.update, after)
        ex = self.assertRaises(exception.ResourceFailure, update_task)
        self.assertIn("Volume resize failed - Unknown status error_extending",
                      str(ex))

        self.assertEqual((rsrc.UPDATE, rsrc.FAILED), rsrc.state)
        self.cinder_fc.volumes.extend.assert_called_once_with(fv.id, 2)

    def test_cinder_volume_extend_attached(self):
        self.stack_name = 'test_cvolume_extend_att_stack'
        self._update_if_attached(self.stack_name)

    def test_cinder_volume_extend_attached_live(self):
        self.patchobject(cinder.CinderClientPlugin, 'get_max_microversion',
                         return_value='3.42')
        self.stack_name = 'test_cvolume_extend_live_att_stack'

        fv = vt_base.FakeVolume('available',
                                size=1, attachments=[])
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     fv, vt_base.FakeVolume('extending'),
                                     vt_base.FakeVolume('extending'),
                                     vt_base.FakeVolume('in-use')])

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume')

        props = copy.deepcopy(rsrc.properties.data)
        props['size'] = 2
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)

        update_task = scheduler.TaskRunner(rsrc.update, after)
        self.assertIsNone(update_task())

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.cinder_fc.volumes.extend.assert_called_once_with(fv.id, 2)
        self.fc.volumes.get_server_volume.assert_not_called()
        self.fc.volumes.delete_server_volume.assert_not_called()

    def test_cinder_volume_extend_created_from_backup_with_same_size(self):
        self.stack_name = 'test_cvolume_extend_snapsht_stack'

        fv = vt_base.FakeVolume('available', size=2)
        self.stub_VolumeBackupConstraint_validate()
        fvbr = vt_base.FakeBackupRestore('vol-123')
        self.patchobject(self.cinder_fc.restores, 'restore', return_value=fvbr)
        self.cinder_fc.volumes.get.side_effect = [
            vt_base.FakeVolume('restoring-backup'),
            vt_base.FakeVolume('available'), fv]
        vol_name = utils.PhysName(self.stack_name, 'volume')
        self.cinder_fc.volumes.update.return_value = None

        self.t['resources']['volume']['properties'] = {
            'availability_zone': 'nova',
            'backup_id': 'backup-123'
        }
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume')
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)
        self.assertEqual('available', fv.status)

        props = copy.deepcopy(rsrc.properties.data)
        props['size'] = 2
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)

        update_task = scheduler.TaskRunner(rsrc.update, after)
        self.assertIsNone(update_task())

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.cinder_fc.restores.restore.assert_called_once_with('backup-123')
        self.cinder_fc.volumes.update.assert_called_once_with(
            'vol-123', description=None, name=vol_name)
        self.assertEqual(3, self.cinder_fc.volumes.get.call_count)

    def test_cinder_volume_retype(self):
        fv = vt_base.FakeVolume('available',
                                size=1, name='my_vol',
                                description='test')
        self.stack_name = 'test_cvolume_retype'
        new_vol_type = 'new_type'
        self.patchobject(cinder.CinderClientPlugin, '_create',
                         return_value=self.cinder_fc)
        self.cinder_fc.volumes.create.return_value = fv
        self.cinder_fc.volumes.get.return_value = fv
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        rsrc = self.create_volume(self.t, stack, 'volume2')

        props = copy.deepcopy(rsrc.properties.data)
        props['volume_type'] = new_vol_type
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)
        self.patchobject(cinder.CinderClientPlugin, 'get_volume_type',
                         return_value=new_vol_type)
        self.patchobject(self.cinder_fc.volumes, 'retype')
        scheduler.TaskRunner(rsrc.update, after)()

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.assertEqual(1, self.cinder_fc.volumes.retype.call_count)

    def test_cinder_volume_update_name_and_metadata(self):
        # update the name, description and metadata
        fv = vt_base.FakeVolume('creating',
                                size=1, name='my_vol',
                                description='test')
        self.stack_name = 'test_cvolume_updname_stack'
        update_name = 'update_name'
        meta = {'Key': 'New Value'}
        update_description = 'update_description'
        kwargs = {
            'name': update_name,
            'description': update_description
        }

        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self._mock_create_volume(fv, self.stack_name,
                                 extra_get_mocks=[fv_ready])
        self.cinder_fc.volumes.update.return_value = None
        self.cinder_fc.volumes.update_all_metadata.return_value = None

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume')

        props = copy.deepcopy(rsrc.properties.data)
        props['name'] = update_name
        props['description'] = update_description
        props['metadata'] = meta
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)
        scheduler.TaskRunner(rsrc.update, after)()

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.cinder_fc.volumes.update.assert_called_once_with(fv_ready,
                                                              **kwargs)
        self.cinder_fc.volumes.update_all_metadata.assert_called_once_with(
            fv_ready, meta)

    def test_cinder_volume_update_read_only(self):
        # update read only access mode
        fv = vt_base.FakeVolume('update_read_only_access_mode')
        self.stack_name = 'test_update_read_only'

        self.cinder_fc.volumes.create.return_value = fv

        update_readonly_mock = self.patchobject(self.cinder_fc.volumes,
                                                'update_readonly_flag',
                                                return_value=None)
        fv_ready = vt_base.FakeVolume('available', id=fv.id, attachments=[])
        self.cinder_fc.volumes.get.return_value = fv_ready

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume')

        props = copy.deepcopy(rsrc.properties.data)
        props['read_only'] = True
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)
        scheduler.TaskRunner(rsrc.update, after)()

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)

        update_readonly_mock.assert_called_once_with(fv.id, True)
        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, availability_zone='nova',
            description='test_description',
            name='test_name',
            metadata={'key': 'value'})
        self.cinder_fc.volumes.get.assert_called_with(fv.id)

    def test_cinder_volume_update_no_need_replace(self):
        # update read only access mode
        fv = vt_base.FakeVolume('creating')
        self.stack_name = 'test_update_no_need_replace'

        self.cinder_fc.volumes.create.return_value = fv

        vol_name = utils.PhysName(self.stack_name, 'volume2')

        fv_ready = vt_base.FakeVolume('available', id=fv.id, size=2,
                                      attachments=[])
        self.cinder_fc.volumes.get.return_value = fv_ready

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume2')

        props = copy.deepcopy(rsrc.properties.data)
        props['size'] = 1
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)
        update_task = scheduler.TaskRunner(rsrc.update, after)
        ex = self.assertRaises(exception.ResourceFailure, update_task)
        self.assertEqual((rsrc.UPDATE, rsrc.FAILED), rsrc.state)
        self.assertIn("NotSupported: resources.volume2: Shrinking volume is "
                      "not supported", str(ex))

        props = copy.deepcopy(rsrc.properties.data)
        props['size'] = 3
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)
        scheduler.TaskRunner(rsrc.update, after)()

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.cinder_fc.volumes.create.assert_called_once_with(
            size=2, availability_zone='nova',
            description=None,
            name=vol_name,
            metadata={}
        )
        self.cinder_fc.volumes.extend.assert_called_once_with(fv.id, 3)
        self.cinder_fc.volumes.get.assert_called_with(fv.id)

    def _update_if_attached(self, stack_name, update_type='resize'):
        # create script
        self.stub_VolumeConstraint_validate()
        fv3 = vt_base.FakeVolume('attaching')
        fv3_ready = vt_base.FakeVolume('in-use', id=fv3.id)
        fv1 = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'),
            extra_create_server_volume_mocks=[fv3])
        attachments = [{'id': 'vol-123',
                        'device': '/dev/vdc',
                        'server_id': 'WikiDatabase'}]
        fv2 = vt_base.FakeVolume('in-use',
                                 attachments=attachments, size=1)
        fvd = vt_base.FakeVolume('in-use')
        resize_m_get = [
            vt_base.FakeVolume('extending'),
            vt_base.FakeVolume('extending'),
            vt_base.FakeVolume('available')
        ]
        extra_get_mocks = [fv1, fv2, fvd, vt_base.FakeVolume('available')]
        if update_type == 'resize':
            extra_get_mocks += resize_m_get
        extra_get_mocks.append(fv3_ready)
        self._mock_create_volume(vt_base.FakeVolume('creating'), stack_name,
                                 extra_get_mocks=extra_get_mocks
                                 )
        # detach script
        self.fc.volumes.get_server_volume.side_effect = [
            fvd, fvd, fakes_nova.fake_exception()]

        if update_type == 'access_mode':
            # update access mode script
            update_readonly_mock = self.patchobject(self.cinder_fc.volumes,
                                                    'update_readonly_flag',
                                                    return_value=None)

        stack = utils.parse_stack(self.t, stack_name=stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume')
        self.create_attachment(self.t, stack, 'attachment')

        props = copy.deepcopy(rsrc.properties.data)
        if update_type == 'access_mode':
            props['read_only'] = True
        if update_type == 'resize':
            props['size'] = 2
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)

        update_task = scheduler.TaskRunner(rsrc.update, after)
        self.assertIsNone(update_task())

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)

        if update_type == 'access_mode':
            update_readonly_mock.assert_called_once_with(fvd.id, True)

        if update_type == 'resize':
            self.cinder_fc.volumes.extend.assert_called_once_with(fvd.id, 2)
        self.fc.volumes.get_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')
        self.fc.volumes.delete_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')

    def test_cinder_volume_update_read_only_attached(self):
        self.stack_name = 'test_cvolume_update_read_only_att_stack'
        self._update_if_attached(self.stack_name, update_type='access_mode')

    def test_cinder_snapshot(self):
        self.stack_name = 'test_cvolume_snpsht_stack'

        self.cinder_fc.volumes.create.return_value = vt_base.FakeVolume(
            'creating')
        fv = vt_base.FakeVolume('available')
        self.cinder_fc.volumes.get.return_value = fv

        fb = vt_base.FakeBackup('creating')
        self.patchobject(self.cinder_fc.backups, 'create', return_value=fb)
        self.patchobject(self.cinder_fc.backups, 'get',
                         return_value=vt_base.FakeBackup('available'))

        t = template_format.parse(single_cinder_volume_template)
        stack = utils.parse_stack(t, stack_name=self.stack_name)

        rsrc = stack['volume']
        self.patchobject(rsrc, '_store_config_default_properties')
        scheduler.TaskRunner(rsrc.create)()

        scheduler.TaskRunner(rsrc.snapshot)()

        self.assertEqual((rsrc.SNAPSHOT, rsrc.COMPLETE), rsrc.state)

        self.assertEqual({'backup_id': 'backup-123'},
                         resource_data_object.ResourceData.get_all(rsrc))
        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, availability_zone=None,
            description='test_description',
            name='test_name',
            metadata={}
        )
        self.cinder_fc.backups.create.assert_called_once_with(fv.id,
                                                              force=True)
        self.cinder_fc.backups.get.assert_called_once_with(fb.id)
        self.cinder_fc.volumes.get.assert_called_with(fv.id)

    def test_cinder_snapshot_error(self):
        self.stack_name = 'test_cvolume_snpsht_err_stack'

        self.cinder_fc.volumes.create.return_value = vt_base.FakeVolume(
            'creating')
        fv = vt_base.FakeVolume('available')
        self.cinder_fc.volumes.get.return_value = fv

        fb = vt_base.FakeBackup('creating')
        self.patchobject(self.cinder_fc.backups, 'create',
                         return_value=fb)
        fail_reason = 'Could not determine which Swift endpoint to use'
        self.patchobject(self.cinder_fc.backups, 'get',
                         return_value=vt_base.FakeBackup(
                             'error', fail_reason=fail_reason))

        t = template_format.parse(single_cinder_volume_template)
        stack = utils.parse_stack(t, stack_name=self.stack_name)

        rsrc = stack['volume']
        self.patchobject(rsrc, '_store_config_default_properties')
        scheduler.TaskRunner(rsrc.create)()

        self.assertRaises(exception.ResourceFailure,
                          scheduler.TaskRunner(rsrc.snapshot))

        self.assertEqual((rsrc.SNAPSHOT, rsrc.FAILED), rsrc.state)
        self.assertIn(fail_reason, rsrc.status_reason)

        self.assertEqual({'backup_id': 'backup-123'},
                         resource_data_object.ResourceData.get_all(rsrc))
        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, availability_zone=None,
            description='test_description',
            name='test_name',
            metadata={}
        )
        self.cinder_fc.backups.create.assert_called_once_with(
            fv.id, force=True)
        self.cinder_fc.volumes.get.assert_called_once_with(fv.id)

    def test_cinder_volume_attachment_update_device(self):
        self.stack_name = 'test_cvolume_attach_udev_stack'

        m_v = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'))
        # attach script
        m_v2 = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'), update=True)

        fva = vt_base.FakeVolume('in-use')
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     m_v, fva,
                                     vt_base.FakeVolume('available'),
                                     m_v2])
        self.stub_VolumeConstraint_validate()

        # delete script
        self.fc.volumes.get_server_volume.side_effect = [
            fva, fva, fakes_nova.fake_exception()]
        self.fc.volumes.delete_server_volume.return_value = None

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        self.create_volume(self.t, stack, 'volume')
        rsrc = self.create_attachment(self.t, stack, 'attachment')

        props = copy.deepcopy(rsrc.properties.data)
        props['mountpoint'] = ''
        props['volume_id'] = 'vol-123'
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)
        scheduler.TaskRunner(rsrc.update, after)()

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.fc.volumes.get_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')
        self.fc.volumes.delete_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')

    def test_cinder_volume_attachment_update_volume(self):

        self.stack_name = 'test_cvolume_attach_uvol_stack'
        fv2 = vt_base.FakeVolume('creating', id='vol-456')
        vol2_name = utils.PhysName(self.stack_name, 'volume2')
        fv3 = vt_base.FakeVolume('available', id=fv2.id)
        fv4 = vt_base.FakeVolume('attaching')
        fv_4ready = vt_base.FakeVolume('in-use', id=fv4.id)
        fv5 = vt_base.FakeVolume('in-use')
        fv6 = vt_base.FakeVolume('available')
        fv7 = vt_base.FakeVolume('attaching', id='vol-456')
        fv7ready = vt_base.FakeVolume('in-use', id=fv7.id)

        self.stub_VolumeConstraint_validate()
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[fv2, fv3, fv_4ready, fv5,
                                                  fv6, fv7ready,
                                                  ],
                                 extra_create_mocks=[fv2])
        self.fc.volumes.create_server_volume.side_effect = [fv4, fv7]
        self.fc.volumes.get_server_volume.side_effect = [
            fv5, fv5, fakes_nova.fake_exception()]
        self.fc.volumes.delete_server_volume.return_value = None

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        self.create_volume(self.t, stack, 'volume')
        self.create_volume(self.t, stack, 'volume2')

        self.fc.volumes.create_server_volume.return_value = fv7

        rsrc = self.create_attachment(self.t, stack, 'attachment')
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)

        props = copy.deepcopy(rsrc.properties.data)
        props['volume_id'] = 'vol-456'
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)
        scheduler.TaskRunner(rsrc.update, after)()

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.assertEqual(fv7.id, rsrc.resource_id)

        self.cinder_fc.volumes.create.assert_called_with(
            size=2, availability_zone='nova',
            description=None,
            name=vol2_name,
            metadata={}
        )
        self.fc.volumes.get_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')
        self.fc.volumes.delete_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')

    def test_cinder_volume_attachment_update_server(self):
        self.stack_name = 'test_cvolume_attach_usrv_stack'

        fv1 = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'))
        fva = vt_base.FakeVolume('in-use')
        fv2 = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'), update=True)
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     fv1, fva,
                                     vt_base.FakeVolume('available'), fv2])
        self.stub_VolumeConstraint_validate()

        # delete script
        self.fc.volumes.get_server_volume.side_effect = [
            fva, fva, fakes_nova.fake_exception()]
        self.fc.volumes.delete_server_volume.return_value = None

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        self.create_volume(self.t, stack, 'volume')

        rsrc = self.create_attachment(self.t, stack, 'attachment')
        self.assertEqual((rsrc.CREATE, rsrc.COMPLETE), rsrc.state)

        props = copy.deepcopy(rsrc.properties.data)
        props['instance_uuid'] = 'AnotherServer'
        props['volume_id'] = 'vol-123'
        after = rsrc_defn.ResourceDefinition(rsrc.name, rsrc.type(), props)
        scheduler.TaskRunner(rsrc.update, after)()

        self.assertEqual((rsrc.UPDATE, rsrc.COMPLETE), rsrc.state)
        self.fc.volumes.get_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')
        self.fc.volumes.delete_server_volume.assert_called_with(
            'WikiDatabase', 'vol-123')

    def test_cinder_volume_attachment_with_serv_resize_task_state(self):
        self.stack_name = 'test_cvolume_attach_usrv_resize_task_state_stack'

        fv1 = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'))
        fva = vt_base.FakeVolume('in-use')
        fv2 = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'), update=True)
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     fv1, fva,
                                     vt_base.FakeVolume('available'), fv2])
        self.stub_VolumeConstraint_validate()
        # delete script
        self.fc.volumes.get_server_volume.side_effect = [
            fva, fva, fakes_nova.fake_exception()]
        self.fc.volumes.delete_server_volume.return_value = None

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        self.create_volume(self.t, stack, 'volume')
        rsrc = self.create_attachment(self.t, stack, 'attachment')
        prg_detach = mock.MagicMock(cinder_complete=True, nova_complete=True)
        prg_attach = mock.MagicMock(called=False, srv_id='InstanceInResize')
        self.fc.volumes.create_server_volume.side_effect = [
            nova_exp.Conflict('409')]

        self.assertEqual(False,
                         rsrc.check_update_complete((prg_detach, prg_attach)))
        self.assertEqual(False, prg_attach.called)

    def test_delete_attachment_has_not_been_created(self):
        self.stack_name = 'test_delete_attachment_has_not_been_created'
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        resource_defn = stack.t.resource_definitions(stack)

        att_rsrc = c_vol.CinderVolumeAttachment(
            'test_attachment',
            resource_defn['attachment'],
            stack)
        att_rsrc.state_set(att_rsrc.UPDATE, att_rsrc.COMPLETE)
        self.assertIsNone(att_rsrc.resource_id)
        # assert even not to create the novaclient instance
        nc = self.patchobject(nova.NovaClientPlugin, '_create')

        scheduler.TaskRunner(att_rsrc.delete)()
        self.assertEqual(0, nc.call_count)
        self.assertEqual((att_rsrc.DELETE, att_rsrc.COMPLETE), att_rsrc.state)

    def test_cinder_create_with_scheduler_hints(self):
        fv = vt_base.FakeVolume('creating')

        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]

        self.stack_name = 'test_cvolume_scheduler_hints_stack'
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        self.patchobject(stack['volume3'], '_store_config_default_properties')
        self.create_volume(self.t, stack, 'volume3')

        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, name='test_name', description=None,
            availability_zone='nova',
            scheduler_hints={'hint1': 'good_advice'},
            metadata={}
        )
        self.assertEqual(2, self.cinder_fc.volumes.get.call_count)

    def test_cinder_create_with_stack_scheduler_hints(self):
        fv = vt_base.FakeVolume('creating')
        sh.cfg.CONF.set_override('stack_scheduler_hints', True)

        self.stack_name = 'test_cvolume_stack_scheduler_hints_stack'
        t = template_format.parse(single_cinder_volume_template)
        stack = utils.parse_stack(t, stack_name=self.stack_name)

        rsrc = stack['volume']

        # rsrc.uuid is only available once the resource has been added.
        stack.add_resource(rsrc)
        self.assertIsNotNone(rsrc.uuid)

        shm = sh.SchedulerHintsMixin
        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]

        self.patchobject(rsrc, '_store_config_default_properties')
        scheduler.TaskRunner(rsrc.create)()
        # this makes sure the auto increment worked on volume creation
        self.assertGreater(rsrc.id, 0)

        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, name='test_name', description='test_description',
            availability_zone=None,
            metadata={},
            scheduler_hints={shm.HEAT_ROOT_STACK_ID: stack.root_stack_id(),
                             shm.HEAT_STACK_ID: stack.id,
                             shm.HEAT_STACK_NAME: stack.name,
                             shm.HEAT_PATH_IN_STACK: [stack.name],
                             shm.HEAT_RESOURCE_NAME: rsrc.name,
                             shm.HEAT_RESOURCE_UUID: rsrc.uuid})
        self.assertEqual(2, self.cinder_fc.volumes.get.call_count)

    def _test_cinder_create_invalid_property_combinations(
            self, stack_name, combinations, err_msg, exc):
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        vp = stack.t['Resources']['volume2']['Properties']
        vp.pop('size')
        vp.update(combinations)
        rsrc = stack['volume2']
        ex = self.assertRaises(exc, rsrc.validate)
        self.assertEqual(err_msg, str(ex))

    def test_cinder_create_with_image_and_imageRef(self):
        self.stack_name = 'test_create_with_image_and_imageRef'
        combinations = {'imageRef': 'image-456', 'image': 'image-123'}
        err_msg = ("Cannot define the following properties at the same time: "
                   "image, imageRef")
        self.stub_ImageConstraint_validate()
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        vp = stack.t['Resources']['volume2']['Properties']
        vp.pop('size')
        vp.update(combinations)
        rsrc = stack.get('volume2')
        ex = self.assertRaises(exception.StackValidationFailed, rsrc.validate)
        self.assertIn(err_msg, str(ex))

    def test_cinder_create_with_image_and_size(self):
        self.stack_name = 'test_create_with_image_and_size'
        combinations = {'image': 'image-123'}
        err_msg = ('If neither "backup_id" nor "size" is provided, one and '
                   'only one of "source_volid", "snapshot_id" must be '
                   'specified, but currently '
                   'specified options: [\'image\'].')
        self.stub_ImageConstraint_validate()
        self._test_cinder_create_invalid_property_combinations(
            self.stack_name, combinations,
            err_msg, exception.StackValidationFailed)

    def test_cinder_create_with_size_snapshot_and_image(self):
        self.stack_name = 'test_create_with_size_snapshot_and_image'
        combinations = {
            'size': 1,
            'image': 'image-123',
            'snapshot_id': 'snapshot-123'}
        self.stub_ImageConstraint_validate()
        self.stub_SnapshotConstraint_validate()
        err_msg = ('If "size" is provided, only one of "image", "imageRef", '
                   '"source_volid", "snapshot_id" can be specified, but '
                   'currently specified options: '
                   '[\'snapshot_id\', \'image\'].')
        self._test_cinder_create_invalid_property_combinations(
            self.stack_name, combinations,
            err_msg, exception.StackValidationFailed)

    def test_cinder_create_with_size_snapshot_and_imageRef(self):
        self.stack_name = 'test_create_with_size_snapshot_and_imageRef'
        combinations = {
            'size': 1,
            'imageRef': 'image-123',
            'snapshot_id': 'snapshot-123'}
        self.stub_ImageConstraint_validate()
        self.stub_SnapshotConstraint_validate()
        # image appears there because of translation rule
        err_msg = ('If "size" is provided, only one of "image", "imageRef", '
                   '"source_volid", "snapshot_id" can be specified, but '
                   'currently specified options: '
                   '[\'snapshot_id\', \'image\'].')
        self._test_cinder_create_invalid_property_combinations(
            self.stack_name, combinations,
            err_msg, exception.StackValidationFailed)

    def test_cinder_create_with_size_snapshot_and_sourcevol(self):
        self.stack_name = 'test_create_with_size_snapshot_and_sourcevol'
        combinations = {
            'size': 1,
            'source_volid': 'volume-123',
            'snapshot_id': 'snapshot-123'}
        self.stub_VolumeConstraint_validate()
        self.stub_SnapshotConstraint_validate()
        err_msg = ('If "size" is provided, only one of "image", "imageRef", '
                   '"source_volid", "snapshot_id" can be specified, but '
                   'currently specified options: '
                   '[\'snapshot_id\', \'source_volid\'].')
        self._test_cinder_create_invalid_property_combinations(
            self.stack_name, combinations,
            err_msg, exception.StackValidationFailed)

    def test_cinder_create_with_snapshot_and_source_volume(self):
        self.stack_name = 'test_create_with_snapshot_and_source_volume'
        combinations = {
            'source_volid': 'source_volume-123',
            'snapshot_id': 'snapshot-123'}
        err_msg = ('If neither "backup_id" nor "size" is provided, one and '
                   'only one of "source_volid", "snapshot_id" must be '
                   'specified, but currently '
                   'specified options: [\'snapshot_id\', \'source_volid\'].')
        self.stub_VolumeConstraint_validate()
        self.stub_SnapshotConstraint_validate()
        self._test_cinder_create_invalid_property_combinations(
            self.stack_name, combinations,
            err_msg, exception.StackValidationFailed)

    def test_cinder_create_with_image_and_source_volume(self):
        self.stack_name = 'test_create_with_image_and_source_volume'
        combinations = {
            'source_volid': 'source_volume-123',
            'image': 'image-123'}
        err_msg = ('If neither "backup_id" nor "size" is provided, one and '
                   'only one of "source_volid", "snapshot_id" must be '
                   'specified, but currently '
                   'specified options: [\'source_volid\', \'image\'].')
        self.stub_VolumeConstraint_validate()
        self.stub_ImageConstraint_validate()
        self._test_cinder_create_invalid_property_combinations(
            self.stack_name, combinations,
            err_msg, exception.StackValidationFailed)

    def test_cinder_create_no_size_no_combinations(self):
        self.stack_name = 'test_create_no_size_no_options'
        combinations = {}
        err_msg = ('If neither "backup_id" nor "size" is provided, one and '
                   'only one of "source_volid", "snapshot_id" must be '
                   'specified, but currently specified options: [].')
        self._test_cinder_create_invalid_property_combinations(
            self.stack_name, combinations,
            err_msg, exception.StackValidationFailed)

    def _test_volume_restore(self, stack_name, final_status='available',
                             stack_final_status=('RESTORE', 'COMPLETE')):
        # create script
        self.cinder_fc.volumes.create.return_value = vt_base.FakeVolume(
            'creating')
        fv = vt_base.FakeVolume('available')
        fv_restoring = vt_base.FakeVolume(
            'restoring-backup', id=fv.id, attachments=[])
        fv_final = vt_base.FakeVolume(final_status, id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_restoring, fv_final]
        self.stub_VolumeBackupConstraint_validate()

        # snapshot script
        fb = vt_base.FakeBackup('creating')
        self.patchobject(self.cinder_fc, 'backups')
        self.cinder_fc.backups.create.return_value = fb
        self.cinder_fc.backups.get.return_value = vt_base.FakeBackup(
            'available')

        # restore script
        fvbr = vt_base.FakeBackupRestore('vol-123')
        self.patchobject(self.cinder_fc.restores, 'restore')
        self.cinder_fc.restores.restore.return_value = fvbr

        t = template_format.parse(single_cinder_volume_template)
        stack = utils.parse_stack(t, stack_name=stack_name)
        self.patchobject(stack['volume'], '_store_config_default_properties')

        scheduler.TaskRunner(stack.create)()

        self.assertEqual((stack.CREATE, stack.COMPLETE), stack.state)

        scheduler.TaskRunner(stack.snapshot, None)()

        self.assertEqual((stack.SNAPSHOT, stack.COMPLETE), stack.state)

        data = stack.prepare_abandon()
        fake_snapshot = collections.namedtuple(
            'Snapshot', ('data', 'stack_id'))(data, stack.id)

        stack.restore(fake_snapshot)

        self.assertEqual(stack_final_status, stack.state)

        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, availability_zone=None,
            description='test_description',
            name='test_name',
            metadata={}
        )
        self.cinder_fc.backups.create.assert_called_once_with(
            fv.id, force=True)
        self.cinder_fc.backups.get.assert_called_once_with(fb.id)
        self.cinder_fc.restores.restore.assert_called_once_with(
            'backup-123', 'vol-123')
        self.assertEqual(3, self.cinder_fc.volumes.get.call_count)

    def test_volume_restore_success(self):
        self._test_volume_restore(stack_name='test_volume_restore_success')

    def test_volume_restore_failed(self):
        self._test_volume_restore(stack_name='test_volume_restore_failed',
                                  final_status='error',
                                  stack_final_status=('RESTORE', 'FAILED'))

    def test_handle_delete_snapshot_no_backup(self):
        self.stack_name = 'test_handle_delete_snapshot_no_backup'
        mock_vs = {
            'resource_data': {}
        }
        t = template_format.parse(single_cinder_volume_template)
        stack = utils.parse_stack(t, stack_name=self.stack_name)
        rsrc = c_vol.CinderVolume(
            'volume',
            stack.t.resource_definitions(stack)['volume'],
            stack)
        self.assertIsNone(rsrc.handle_delete_snapshot(mock_vs))

    def test_vaildate_deletion_policy(self):
        cfg.CONF.set_override('backups_enabled', False, group='volumes')
        self.stack_name = 'test_volume_validate_deletion_policy'
        self.t['resources']['volume']['deletion_policy'] = 'Snapshot'
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        rsrc = self.get_volume(self.t, stack, 'volume')
        self.assertRaisesRegex(
            exception.StackValidationFailed,
            'volume backup service is not enabled',
            rsrc.validate)

    def test_volume_get_live_state(self):
        tmpl = """
        heat_template_version: 2013-05-23
        description: Cinder volume
        resources:
          volume:
            type: OS::Cinder::Volume
            properties:
              size: 1
              name: test_name
              description: test_description
              image: 1234
              scheduler_hints:
                'consistencygroup_id': 4444
        """
        t = template_format.parse(tmpl)
        stack = utils.parse_stack(t, stack_name='get_live_state')
        rsrc = stack['volume']
        rsrc._availability_zone = 'nova'
        rsrc.resource_id = '1234'

        vol_resp = {
            'attachments': [],
            'availability_zone': 'nova',
            'snapshot_id': None,
            'size': 1,
            'metadata': {'test': 'test_value', 'readonly': False},
            'consistencygroup_id': '4444',
            'volume_image_metadata': {'image_id': '1234',
                                      'image_name': 'test'},
            'description': None,
            'source_volid': None,
            'name': 'test-volume-jbdbgdsy3vyg',
            'volume_type': 'lvmdriver-1'
        }
        vol = mock.MagicMock()
        vol.to_dict.return_value = vol_resp
        rsrc.client().volumes = mock.MagicMock()
        rsrc.client().volumes.get = mock.MagicMock(return_value=vol)
        rsrc.client().volume_api_version = 3
        rsrc.data = mock.MagicMock(return_value={'volume_type': 'lvmdriver-1'})

        reality = rsrc.get_live_state(rsrc.properties)
        expected = {
            'size': 1,
            'metadata': {'test': 'test_value'},
            'description': None,
            'name': 'test-volume-jbdbgdsy3vyg',
            'backup_id': None,
            'read_only': False,
        }

        self.assertEqual(expected, reality)

    def test_detach_volume_to_complete_with_resize_task_state(self):
        fv = vt_base.FakeVolume('creating')
        self.stack_name = 'test_cvolume_detach_with_resize_task_state_stack'

        self.stub_SnapshotConstraint_validate()
        self.stub_VolumeConstraint_validate()
        self.stub_VolumeTypeConstraint_validate()
        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]
        self.fc.volumes.delete_server_volume.side_effect = [
            nova_exp.Conflict('409')]
        self.t['resources']['volume']['properties'].update({
            'volume_type': 'lvm',
        })
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        rsrc = self.create_volume(self.t, stack, 'volume')
        prg_detach = mock.MagicMock(called=False, srv_id='InstanceInResize')
        self.assertEqual(False, rsrc._detach_volume_to_complete(prg_detach))
        self.assertEqual(False, prg_detach.called)

    def test_detach_volume_to_complete_with_active_task_state(self):
        fv = vt_base.FakeVolume('creating')
        self.stack_name = 'test_cvolume_detach_with_active_task_state_stack'

        self.stub_SnapshotConstraint_validate()
        self.stub_VolumeConstraint_validate()
        self.stub_VolumeTypeConstraint_validate()
        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]

        self.t['resources']['volume']['properties'].update({
            'volume_type': 'lvm',
        })
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        rsrc = self.create_volume(self.t, stack, 'volume')
        prg_detach = mock.MagicMock(called=False, srv_id='InstanceInActive')
        self.assertEqual(False, rsrc._detach_volume_to_complete(prg_detach))
        self.assertEqual(True, prg_detach.called)

    def test_attach_volume_to_complete_with_resize_task_state(self):
        fv = vt_base.FakeVolume('creating')
        self.stack_name = 'test_cvolume_attach_with_resize_task_state_stack'

        self.stub_SnapshotConstraint_validate()
        self.stub_VolumeConstraint_validate()
        self.stub_VolumeTypeConstraint_validate()
        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]
        self.fc.volumes.create_server_volume.side_effect = [
            nova_exp.Conflict('409')]
        self.t['resources']['volume']['properties'].update({
            'volume_type': 'lvm',
        })
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        rsrc = self.create_volume(self.t, stack, 'volume')
        prg_attach = mock.MagicMock(called=False, srv_id='InstanceInResize')
        self.assertEqual(False, rsrc._attach_volume_to_complete(prg_attach))
        self.assertEqual(False, prg_attach.called)

    def test_attach_volume_to_complete_with_active_task_state(self):
        fv = vt_base.FakeVolume('creating')
        self.stack_name = 'test_cvolume_attach_with_active_task_state_stack'

        self.stub_SnapshotConstraint_validate()
        self.stub_VolumeConstraint_validate()
        self.stub_VolumeTypeConstraint_validate()
        self.cinder_fc.volumes.create.return_value = fv
        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]

        self.t['resources']['volume']['properties'].update({
            'volume_type': 'lvm',
        })
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        rsrc = self.create_volume(self.t, stack, 'volume')
        self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'))

        prg_attach = mock.MagicMock(called=False, srv_id='InstanceInActive')
        self.assertEqual(False, rsrc._attach_volume_to_complete(prg_attach))
        self.assertEqual('vol-123', prg_attach.called)

    def test_empty_string_az(self):
        fv = vt_base.FakeVolume('creating')
        self.stack_name = 'test_cvolume_default_stack'

        vol_name = utils.PhysName(self.stack_name, 'volume')
        self.cinder_fc.volumes.create.return_value = fv
        fv_ready = vt_base.FakeVolume('available', id=fv.id)
        self.cinder_fc.volumes.get.side_effect = [fv, fv_ready]

        self.t['resources']['volume']['properties'] = {
            'size': '1',
            'availability_zone': "",
        }
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        self.create_volume(self.t, stack, 'volume')

        self.cinder_fc.volumes.create.assert_called_once_with(
            size=1, availability_zone=None,
            description=None,
            name=vol_name,
            metadata={}
        )

    def test_ready_to_extend_volume(self):
        self.stack_name = 'test_ready_to_extend_volume'

        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     vt_base.FakeVolume('extending'),
                                     vt_base.FakeVolume('reserved'),
                                     vt_base.FakeVolume('in-use'),
                                     vt_base.FakeVolume('available'),
                                     vt_base.FakeVolume('creating'),
                                     vt_base.FakeVolume('extending'),
                                     vt_base.FakeVolume('reserved'),
                                     vt_base.FakeVolume('in-use'),
                                     vt_base.FakeVolume('available')])

        stack = utils.parse_stack(self.t, stack_name=self.stack_name)

        rsrc = self.create_volume(self.t, stack, 'volume',
                                  mock_check_extend_ready=False)

        self.assertEqual(False, rsrc._ready_to_extend_volume())
        self.assertEqual(False, rsrc._ready_to_extend_volume())
        self.assertEqual(False, rsrc._ready_to_extend_volume())
        self.assertEqual(True, rsrc._ready_to_extend_volume())

        self.patchobject(cinder.CinderClientPlugin, 'get_max_microversion',
                         return_value='3.42')
        self.assertEqual(False, rsrc._ready_to_extend_volume())
        self.assertEqual(False, rsrc._ready_to_extend_volume())
        self.assertEqual(False, rsrc._ready_to_extend_volume())
        self.assertEqual(True, rsrc._ready_to_extend_volume())
        self.assertEqual(True, rsrc._ready_to_extend_volume())

    def test_try_detach_volume_if_server_was_temporarily_in_error(self):
        self.stack_name = 'test_cvolume_detach_server_in_error_stack'
        fva = vt_base.FakeVolume('in-use')
        m_v = self._mock_create_server_volume_script(
            vt_base.FakeVolume('attaching'))
        self._mock_create_volume(vt_base.FakeVolume('creating'),
                                 self.stack_name,
                                 extra_get_mocks=[
                                     m_v, fva,
                                     vt_base.FakeVolume('available')])
        self.stub_VolumeConstraint_validate()
        # delete script
        self.fc.volumes.get_server_volume.side_effect = [
            fva, fva, fakes_nova.fake_exception()]
        nova_responses = [nova_exp.Conflict('409'), None]
        self.fc.volumes.delete_server_volume.side_effect = nova_responses
        stack = utils.parse_stack(self.t, stack_name=self.stack_name)
        self.create_volume(self.t, stack, 'volume')
        rsrc = self.create_attachment(self.t, stack, 'attachment')
        scheduler.TaskRunner(rsrc.delete)()
        self.assertEqual(self.fc.volumes.delete_server_volume.call_count,
                         len(nova_responses))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

as_params = {'KeyName': 'test', 'ImageId': 'foo'}
as_template = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Description" : "AutoScaling Test",
  "Parameters" : {
  "ImageId": {"Type": "String"},
  "KeyName": {"Type": "String"}
  },
  "Resources" : {
    "WebServerGroup" : {
      "Type" : "AWS::AutoScaling::AutoScalingGroup",
      "Properties" : {
        "AvailabilityZones" : ["nova"],
        "LaunchConfigurationName" : { "Ref" : "LaunchConfig" },
        "MinSize" : "1",
        "MaxSize" : "5",
        "LoadBalancerNames" : [ { "Ref" : "ElasticLoadBalancer" } ]
      }
    },
    "WebServerScaleUpPolicy" : {
      "Type" : "AWS::AutoScaling::ScalingPolicy",
      "Properties" : {
        "AdjustmentType" : "ChangeInCapacity",
        "AutoScalingGroupName" : { "Ref" : "WebServerGroup" },
        "Cooldown" : "60",
        "ScalingAdjustment" : "1"
      }
    },
    "WebServerScaleDownPolicy" : {
      "Type" : "AWS::AutoScaling::ScalingPolicy",
      "Properties" : {
        "AdjustmentType" : "ChangeInCapacity",
        "AutoScalingGroupName" : { "Ref" : "WebServerGroup" },
        "Cooldown" : "60",
        "ScalingAdjustment" : "-1"
      }
    },
    "ElasticLoadBalancer" : {
        "Type" : "AWS::ElasticLoadBalancing::LoadBalancer",
        "Properties" : {
            "AvailabilityZones" : ["nova"],
            "Listeners" : [ {
                "LoadBalancerPort" : "80",
                "InstancePort" : "80",
                "Protocol" : "HTTP"
            }]
        }
    },
    "LaunchConfig" : {
      "Type" : "AWS::AutoScaling::LaunchConfiguration",
      "Properties": {
        "ImageId" : {"Ref": "ImageId"},
        "InstanceType"   : "bar",
        "BlockDeviceMappings": [
            {
                "DeviceName": "vdb",
                "Ebs": {"SnapshotId": "9ef5496e-7426-446a-bbc8-01f84d9c9972",
                        "DeleteOnTermination": "True"}
            }]
      }
    }
  }
}
'''

as_heat_template = '''
    heat_template_version: 2013-05-23
    description: AutoScaling Test
    resources:
      my-group:
        type: OS::Heat::AutoScalingGroup
        properties:
          max_size: 5
          min_size: 1
          resource:
            type: ResourceWithPropsAndAttrs
            properties:
                Foo: hello
      my-policy:
        type: OS::Heat::ScalingPolicy
        properties:
          auto_scaling_group_id: {get_resource: my-group}
          scaling_adjustment: 1
          adjustment_type: change_in_capacity
          cooldown: 60
'''

as_template_bad_group = '''
{
  "AWSTemplateFormatVersion" : "2010-09-09",
  "Parameters" : {
  "ImageId": {"Type": "String"},
  "KeyName": {"Type": "String"}
  },
  "Resources" : {
    "WebServerScaleUpPolicy" : {
      "Type" : "AWS::AutoScaling::ScalingPolicy",
      "Properties" : {
        "AdjustmentType" : "ChangeInCapacity",
        "AutoScalingGroupName" : "not a real group",
        "Cooldown" : "60",
        "ScalingAdjustment" : "1"
      }
    }
  }
}
'''

as_heat_template_bad_group = '''
    heat_template_version: 2013-05-23
    description: AutoScaling Test
    resources:
      my-policy:
        type: OS::Heat::ScalingPolicy
        properties:
          auto_scaling_group_id: bad-group
          scaling_adjustment: 1
          adjustment_type: change_in_capacity
          cooldown: 60
'''



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import collections
import copy
import functools
import hashlib

from stevedore import extension

from heat.common import exception
from heat.common.i18n import _
from heat.common import template_format
from heat.engine import conditions
from heat.engine import environment
from heat.engine import function
from heat.engine import template_files
from heat.objects import raw_template as template_object

__all__ = ['Template']


_template_classes = None


def get_version(template_data, available_versions):
    version_keys = set(key for key, version in available_versions)
    candidate_keys = set(k for k, v in template_data.items() if
                         isinstance(v, str))

    keys_present = version_keys & candidate_keys

    if len(keys_present) > 1:
        explanation = _('Ambiguous versions (%s)') % ', '.join(keys_present)
        raise exception.InvalidTemplateVersion(explanation=explanation)
    try:
        version_key = keys_present.pop()
    except KeyError:
        explanation = _('Template version was not provided')
        raise exception.InvalidTemplateVersion(explanation=explanation)
    return version_key, template_data[version_key]


def _get_template_extension_manager():
    return extension.ExtensionManager(
        namespace='heat.templates',
        invoke_on_load=False,
        on_load_failure_callback=raise_extension_exception)


def raise_extension_exception(extmanager, ep, err):
    raise TemplatePluginNotRegistered(name=ep.name, error=str(err))


class TemplatePluginNotRegistered(exception.HeatException):
    msg_fmt = _("Could not load %(name)s: %(error)s")


def get_template_class(template_data):
    available_versions = _template_classes.keys()
    version = get_version(template_data, available_versions)
    version_type = version[0]
    try:
        return _template_classes[version]
    except KeyError:
        av_list = sorted(
            [v for k, v in available_versions if k == version_type])
        msg_data = {'version': ': '.join(version),
                    'version_type': version_type,
                    'available': ', '.join(v for v in av_list)}

        if len(av_list) > 1:
            explanation = _('"%(version)s". "%(version_type)s" '
                            'should be one of: %(available)s') % msg_data
        else:
            explanation = _('"%(version)s". "%(version_type)s" '
                            'should be: %(available)s') % msg_data
        raise exception.InvalidTemplateVersion(explanation=explanation)


class Template(collections.abc.Mapping):
    """Abstract base class for template format plugins.

    All template formats (both internal and third-party) should derive from
    Template and implement the abstract functions to provide resource
    definitions and other data.

    This is a stable third-party API. Do not add implementations that are
    specific to internal template formats. Do not add new abstract methods.
    """

    condition_functions = {}
    functions = {}

    def __new__(cls, template, *args, **kwargs):
        """Create a new Template of the appropriate class."""
        global _template_classes

        if _template_classes is None:
            mgr = _get_template_extension_manager()
            _template_classes = dict((tuple(name.split('.')), mgr[name].plugin)
                                     for name in mgr.names())

        if cls != Template:
            TemplateClass = cls
        else:
            TemplateClass = get_template_class(template)

        return super(Template, cls).__new__(TemplateClass)

    def __init__(self, template, template_id=None, files=None, env=None):
        """Initialise the template with JSON object and set of Parameters."""
        self.id = template_id
        self.t = template
        self.files = files or {}
        self.maps = self[self.MAPPINGS]
        self.env = env or environment.Environment({})
        self.merge_sections = [self.PARAMETERS]

        self.version = get_version(self.t, _template_classes.keys())
        self.t_digest = None

        condition_functions = {n: function.Invalid for n in self.functions}
        condition_functions.update(self.condition_functions)
        self._parser_condition_functions = condition_functions

    def __deepcopy__(self, memo):
        return Template(copy.deepcopy(self.t, memo), files=self.files,
                        env=self.env)

    def merge_snippets(self, other):
        for s in self.merge_sections:
            if s not in other.t:
                continue
            if s not in self.t:
                self.t[s] = {}
            self.t[s].update(other.t[s])

    @classmethod
    def load(cls, context, template_id, t=None):
        """Retrieve a Template with the given ID from the database."""
        if t is None:
            t = template_object.RawTemplate.get_by_id(context, template_id)
        env = environment.Environment(t.environment)
        # support loading the legacy t.files, but modern templates will
        # have a t.files_id
        t_files = t.files or t.files_id
        return cls(t.template, template_id=template_id, env=env,
                   files=t_files)

    def store(self, context):
        """Store the Template in the database and return its ID."""
        rt = {
            'template': self.t,
            'files_id': self.files.store(context),
            'environment': self.env.env_as_dict()
        }
        if self.id is None:
            new_rt = template_object.RawTemplate.create(context, rt)
            self.id = new_rt.id
        else:
            template_object.RawTemplate.update_by_id(context, self.id, rt)
        return self.id

    @property
    def files(self):
        return self._template_files

    @files.setter
    def files(self, files):
        self._template_files = template_files.TemplateFiles(files)

    def __iter__(self):
        """Return an iterator over the section names."""
        return (s for s in self.SECTIONS
                if s not in self.SECTIONS_NO_DIRECT_ACCESS)

    def __len__(self):
        """Return the number of sections."""
        return len(self.SECTIONS) - len(self.SECTIONS_NO_DIRECT_ACCESS)

    @abc.abstractmethod
    def param_schemata(self, param_defaults=None):
        """Return a dict of parameters.Schema objects for the parameters."""
        pass

    def all_param_schemata(self, files):
        schema = {}
        files = files if files is not None else {}
        for f in files.values():
            try:
                data = template_format.parse(f)
            except ValueError:
                continue
            else:
                sub_tmpl = Template(data)
                schema.update(sub_tmpl.param_schemata())
        # Parent template has precedence, so update the schema last.
        schema.update(self.param_schemata())
        return schema

    @abc.abstractmethod
    def get_section_name(self, section):
        """Get the name of a field within a resource or output definition.

        Return the name of the given field (specified by the constants given
        in heat.engine.rsrc_defn and heat.engine.output) in the template
        format. This is used in error reporting to help users find the
        location of errors in the template.

        Note that 'section' here does not refer to a top-level section of the
        template (like parameters, resources, &c.) as it does everywhere else.
        """
        pass

    @abc.abstractmethod
    def parameters(self, stack_identifier, user_params, param_defaults=None):
        """Return a parameters.Parameters object for the stack."""
        pass

    def validate_resource_definitions(self, stack):
        """Check validity of resource definitions.

        This method is deprecated. Subclasses should validate the resource
        definitions in the process of generating them when calling
        resource_definitions(). However, for now this method is still called
        in case any third-party plugins are relying on this for validation and
        need time to migrate.
        """
        pass

    def conditions(self, stack):
        """Return a dictionary of resolved conditions."""
        return conditions.Conditions({})

    @abc.abstractmethod
    def outputs(self, stack):
        """Return a dictionary of OutputDefinition objects."""
        pass

    @abc.abstractmethod
    def resource_definitions(self, stack):
        """Return a dictionary of ResourceDefinition objects."""
        pass

    @abc.abstractmethod
    def add_resource(self, definition, name=None):
        """Add a resource to the template.

        The resource is passed as a ResourceDefinition object. If no name is
        specified, the name from the ResourceDefinition should be used.
        """
        pass

    def add_output(self, definition):
        """Add an output to the template.

        The output is passed as a OutputDefinition object.
        """
        raise NotImplementedError

    def remove_resource(self, name):
        """Remove a resource from the template."""
        self.t.get(self.RESOURCES, {}).pop(name)

    def remove_all_resources(self):
        """Remove all the resources from the template."""
        if self.RESOURCES in self.t:
            self.t.update({self.RESOURCES: {}})

    def parse(self, stack, snippet, path=''):
        return parse(self.functions, stack, snippet, path, self)

    def parse_condition(self, stack, snippet, path=''):
        return parse(self._parser_condition_functions, stack, snippet,
                     path, self)

    def validate(self):
        """Validate the template.

        Validates the top-level sections of the template as well as syntax
        inside select sections. Some sections are not checked here but in
        code parts that are responsible for working with the respective
        sections (e.g. parameters are check by parameters schema class).
        """
        t_digest = hashlib.sha256(
            str(self.t).encode('utf-8')).hexdigest()

        # TODO(kanagaraj-manickam) currently t_digest is stored in self. which
        # is used to check whether already template is validated or not.
        # But it needs to be loaded from dogpile cache backend once its
        # available in heat (https://specs.openstack.org/openstack/heat-specs/
        # specs/liberty/constraint-validation-cache.html). This is required
        # as multiple heat-engines may process the same template at least
        # in case of instance_group. And it fixes partially bug 1444316

        if t_digest == self.t_digest:
            return

        # check top-level sections
        for k in self.t.keys():
            if k not in self.SECTIONS:
                raise exception.InvalidTemplateSection(section=k)

        # check resources
        for res in self[self.RESOURCES].values():
            try:
                if not res or not res.get('Type'):
                    message = _('Each Resource must contain '
                                'a Type key.')
                    raise exception.StackValidationFailed(message=message)
            except AttributeError:
                message = _('Resources must contain Resource. '
                            'Found a [%s] instead') % type(res)
                raise exception.StackValidationFailed(message=message)
        self.t_digest = t_digest

    @classmethod
    def create_empty_template(cls,
                              version=('heat_template_version', '2015-04-30'),
                              from_template=None):
        """Create an empty template.

        Creates a new empty template with given version. If version is
        not provided, a new empty HOT template of version "2015-04-30"
        is returned.

        :param version: A tuple containing version header of the template
                        version key and value,
                        e.g. ``('heat_template_version', '2015-04-30')``
        :returns: A new empty template.
        """
        if from_template:
            # remove resources from the template and return; keep the
            # env and other things intact
            tmpl = copy.deepcopy(from_template)
            tmpl.remove_all_resources()
            return tmpl
        else:
            tmpl = {version[0]: version[1]}
            return cls(tmpl)


def parse(functions, stack, snippet, path='', template=None):
    recurse = functools.partial(parse, functions, stack, template=template)

    if isinstance(snippet, collections.abc.Mapping):
        def mkpath(key):
            return '.'.join([path, str(key)])

        if len(snippet) == 1:
            fn_name, args = next(iter(snippet.items()))
            Func = functions.get(fn_name)
            if Func is not None:
                try:
                    path = '.'.join([path, fn_name])
                    if (isinstance(Func, type) and
                            issubclass(Func, function.Macro)):
                        return Func(stack, fn_name, args,
                                    functools.partial(recurse, path=path),
                                    template)
                    else:
                        return Func(stack, fn_name, recurse(args, path))
                except (ValueError, TypeError, KeyError) as e:
                    raise exception.StackValidationFailed(
                        path=path,
                        message=str(e))

        return dict((k, recurse(v, mkpath(k)))
                    for k, v in snippet.items())
    elif (not isinstance(snippet, str) and
          isinstance(snippet, collections.abc.Iterable)):

        def mkpath(idx):
            return ''.join([path, '[%d]' % idx])

        return [recurse(v, mkpath(i)) for i, v in enumerate(snippet)]
    else:
        return snippet



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine.resources.openstack.heat import none_resource
from heat.engine import support


class GlanceWebImage(resource.Resource):
    """A resource managing images in Glance using web-download import.

    This provides image support for recent Glance installation.
    """

    support_status = support.SupportStatus(version='12.0.0')

    PROPERTIES = (
        NAME, IMAGE_ID, MIN_DISK, MIN_RAM, PROTECTED,
        DISK_FORMAT, CONTAINER_FORMAT, LOCATION, TAGS,
        ARCHITECTURE, KERNEL_ID, OS_DISTRO, OS_VERSION, OWNER,
        EXTRA_PROPERTIES, VISIBILITY, RAMDISK_ID, ACTIVE, MEMBERS
    ) = (
        'name', 'id', 'min_disk', 'min_ram', 'protected',
        'disk_format', 'container_format', 'location', 'tags',
        'architecture', 'kernel_id', 'os_distro', 'os_version',
        'owner', 'extra_properties', 'visibility', 'ramdisk_id',
        'active', 'members'
    )

    glance_id_pattern = ('^([0-9a-fA-F]){8}-([0-9a-fA-F]){4}-([0-9a-fA-F]){4}'
                         '-([0-9a-fA-F]){4}-([0-9a-fA-F]){12}$')

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name for the image. The name of an image is not '
              'unique to a Image Service node.')
        ),
        IMAGE_ID: properties.Schema(
            properties.Schema.STRING,
            _('The image ID. Glance will generate a UUID if not specified.')
        ),
        MIN_DISK: properties.Schema(
            properties.Schema.INTEGER,
            _('Amount of disk space (in GB) required to boot image. '
              'Default value is 0 if not specified '
              'and means no limit on the disk size.'),
            constraints=[
                constraints.Range(min=0),
            ],
            default=0
        ),
        MIN_RAM: properties.Schema(
            properties.Schema.INTEGER,
            _('Amount of ram (in MB) required to boot image. Default value '
              'is 0 if not specified and means no limit on the ram size.'),
            constraints=[
                constraints.Range(min=0),
            ],
            default=0
        ),
        PROTECTED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Whether the image can be deleted. If the value is True, '
              'the image is protected and cannot be deleted.'),
            update_allowed=True,
            default=False
        ),
        DISK_FORMAT: properties.Schema(
            properties.Schema.STRING,
            _('Disk format of image.'),
            required=True,
            constraints=[
                constraints.AllowedValues(
                    ['ami', 'ari', 'aki', 'vhd', 'vhdx', 'vmdk', 'raw',
                     'qcow2', 'vdi', 'iso', 'ploop'])
            ]
        ),
        CONTAINER_FORMAT: properties.Schema(
            properties.Schema.STRING,
            _('Container format of image.'),
            required=True,
            constraints=[
                constraints.AllowedValues([
                    'ami', 'ari', 'aki', 'bare', 'ovf', 'ova', 'docker'])
            ]
        ),
        LOCATION: properties.Schema(
            properties.Schema.STRING,
            _('URL where the data for this image already resides. For '
              'example, if the image data is stored in swift, you could '
              'specify "swift://example.com/container/obj".'),
            required=True,
        ),
        TAGS: properties.Schema(
            properties.Schema.LIST,
            _('List of image tags.'),
            update_allowed=True,
        ),
        ARCHITECTURE: properties.Schema(
            properties.Schema.STRING,
            _('Operating system architecture.'),
            update_allowed=True,
        ),
        KERNEL_ID: properties.Schema(
            properties.Schema.STRING,
            _('ID of image stored in Glance that should be used as '
              'the kernel when booting an AMI-style image.'),
            update_allowed=True,
            constraints=[
                constraints.AllowedPattern(glance_id_pattern)
            ]
        ),
        OS_DISTRO: properties.Schema(
            properties.Schema.STRING,
            _('The common name of the operating system distribution '
              'in lowercase.'),
            update_allowed=True,
        ),
        OS_VERSION: properties.Schema(
            properties.Schema.STRING,
            _('Operating system version as specified by the distributor.'),
            update_allowed=True,
        ),
        OWNER: properties.Schema(
            properties.Schema.STRING,
            _('Owner of the image.'),
            update_allowed=True,
        ),
        EXTRA_PROPERTIES: properties.Schema(
            properties.Schema.MAP,
            _('Arbitrary properties to associate with the image.'),
            update_allowed=True,
            default={},
            support_status=support.SupportStatus(version='17.0.0')
        ),
        VISIBILITY: properties.Schema(
            properties.Schema.STRING,
            _('Scope of image accessibility.'),
            update_allowed=True,
            default='private',
            constraints=[
                constraints.AllowedValues(
                    ['public', 'private', 'community', 'shared'])
            ]
        ),
        RAMDISK_ID: properties.Schema(
            properties.Schema.STRING,
            _('ID of image stored in Glance that should be used as '
              'the ramdisk when booting an AMI-style image.'),
            update_allowed=True,
            constraints=[
                constraints.AllowedPattern(glance_id_pattern)
            ]
        ),
        ACTIVE: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Activate or deactivate the image. Requires Admin Access.'),
            default=True,
            update_allowed=True,
            support_status=support.SupportStatus(version='16.0.0')
        ),
        MEMBERS: properties.Schema(
            properties.Schema.LIST,
            _('List of additional members that are permitted '
              'to read the image. This may be a Keystone Project '
              'IDs or User IDs, depending on the Glance configuration '
              'in use.'),
            schema=properties.Schema(
                properties.Schema.STRING,
                _('A member ID. This may be a Keystone Project ID '
                  'or User ID, depending on the Glance configuration '
                  'in use.')
            ),
            update_allowed=True,
            support_status=support.SupportStatus(version='16.0.0')
        )
    }

    default_client_name = 'glance'

    entity = 'images'

    def handle_create(self):
        args = dict((k, v) for k, v in self.properties.items()
                    if v is not None and k is not self.EXTRA_PROPERTIES)
        members = args.pop(self.MEMBERS, [])
        active = args.pop(self.ACTIVE)
        location = args.pop(self.LOCATION)
        images = self.client().images
        image = images.create(**args)
        image_id = image.id
        self.resource_id_set(image_id)
        images.image_import(image_id, method='web-download', uri=location)
        for member in members:
            self.client().image_members.create(image_id, member)
        props = self.properties.get(self.EXTRA_PROPERTIES)
        images.update(image.id, **props)
        return active

    def check_create_complete(self, active):
        image = self.client().images.get(self.resource_id)
        if image.status == 'killed':
            raise exception.ResourceInError(
                resource_status=image.status,
            )
        if not active:
            if image.status == 'active':
                self.client().images.deactivate(self.resource_id)
                return True
            elif image.status == 'deactivated':
                return True
            else:
                return False
        else:
            return image.status == 'active'

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        images = self.client().images
        if prop_diff:
            active = prop_diff.pop(self.ACTIVE, None)
            if active is False:
                images.deactivate(self.resource_id)

            if self.TAGS in prop_diff:
                existing_tags = self.properties.get(self.TAGS) or []
                diff_tags = prop_diff.pop(self.TAGS) or []

                new_tags = set(diff_tags) - set(existing_tags)
                for tag in new_tags:
                    self.client().image_tags.update(
                        self.resource_id,
                        tag)

                removed_tags = set(existing_tags) - set(diff_tags)
                for tag in removed_tags:
                    with self.client_plugin().ignore_not_found:
                        self.client().image_tags.delete(
                            self.resource_id,
                            tag)

        if self.EXTRA_PROPERTIES in prop_diff:
            old_properties = self.properties.get(self.EXTRA_PROPERTIES)
            new_properties = prop_diff.pop(self.EXTRA_PROPERTIES)
            prop_diff.update(new_properties)
            remove_props = list(set(old_properties) - set(new_properties))

            # Though remove_props defaults to None within the glanceclient,
            # setting it to a list (possibly []) every time ensures only one
            # calling format to images.update
            images.update(self.resource_id, remove_props, **prop_diff)
        else:
            images.update(self.resource_id, **prop_diff)

            if self.MEMBERS in prop_diff:
                existing_members = self.properties.get(self.MEMBERS) or []
                diff_members = prop_diff.pop(self.MEMBERS) or []

                new_members = set(diff_members) - set(existing_members)
                for _member in new_members:
                    self.glance().image_members.create(
                        self.resource_id, _member)
                removed_members = set(existing_members) - set(diff_members)
                for _member in removed_members:
                    self.glance().image_members.delete(
                        self.resource_id, _member)

        return active

    def check_update_complete(self, active):
        if active:
            self.client().images.reactivate(self.resource_id)
        return True

    def validate(self):
        super(GlanceWebImage, self).validate()
        container_format = self.properties[self.CONTAINER_FORMAT]
        if (container_format in ['ami', 'ari', 'aki']
                and self.properties[self.DISK_FORMAT] != container_format):
            msg = _("Invalid mix of disk and container formats. When "
                    "setting a disk or container format to one of 'aki', "
                    "'ari', or 'ami', the container and disk formats must "
                    "match.")
            raise exception.StackValidationFailed(message=msg)

        if (self.properties[self.MEMBERS]
                and self.properties[self.VISIBILITY] != 'shared'):
            raise exception.ResourcePropertyValueDependency(
                prop1=self.MEMBERS,
                prop2=self.VISIBILITY,
                value='shared')

    def get_live_resource_data(self):
        image_data = super(GlanceWebImage, self).get_live_resource_data()
        if image_data.get('status') in ('deleted', 'killed'):
            raise exception.EntityNotFound(entity='Resource',
                                           name=self.name)
        return image_data

    def parse_live_resource_data(self, resource_properties, resource_data):
        image_reality = {}

        for key in self.PROPERTIES:
            if key == self.LOCATION:
                continue
            if key == self.IMAGE_ID:
                if (resource_properties.get(self.IMAGE_ID) is not None or
                        resource_data.get(self.IMAGE_ID) != self.resource_id):
                    image_reality.update({self.IMAGE_ID: resource_data.get(
                        self.IMAGE_ID)})
                else:
                    image_reality.update({self.IMAGE_ID: None})

            if key == self.EXTRA_PROPERTIES:
                continue
            else:
                image_reality.update({key: resource_data.get(key)})

        if resource_properties.get(self.EXTRA_PROPERTIES):
            extra_properties = {}
            for key in resource_properties.get(self.EXTRA_PROPERTIES):
                extra_properties[key] = resource_data.get(key)
            image_reality.update({self.EXTRA_PROPERTIES: extra_properties})

        return image_reality


class GlanceImage(none_resource.NoneResource):
    """A resource managing images in Glance.

    A resource provides managing images that are meant to be used with other
    services.
    """

    support_status = support.SupportStatus(
        status=support.HIDDEN,
        version='22.0.0',
        previous_status=support.SupportStatus(
            status=support.DEPRECATED,
            version='8.0.0',
            message=_(
                'Creating a Glance Image based on an existing URL location '
                'requires the Glance v1 API, which is deprecated.'),
            previous_status=support.SupportStatus(version='2014.2')
        )
    )


def resource_mapping():
    return {
        'OS::Glance::Image': GlanceImage,
        'OS::Glance::WebImage': GlanceWebImage
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import yaml

from heat.common.i18n import _
from heat.common import template_format


SECTIONS = (
    PARAMETERS, RESOURCE_REGISTRY, PARAMETER_DEFAULTS,
    ENCRYPTED_PARAM_NAMES, EVENT_SINKS,
    PARAMETER_MERGE_STRATEGIES,
) = (
    'parameters', 'resource_registry', 'parameter_defaults',
    'encrypted_param_names', 'event_sinks',
    'parameter_merge_strategies',
)


def parse(env_str):
    """Takes a string and returns a dict containing the parsed structure."""
    if env_str is None:
        return {}

    try:
        env = template_format.yaml.load(env_str,
                                        Loader=template_format.yaml_loader)
    except yaml.YAMLError:
        # NOTE(prazumovsky): we need to return more informative error for
        # user, so use SafeLoader, which return error message with template
        # snippet where error has been occurred.
        try:
            env = yaml.load(env_str, Loader=yaml.SafeLoader)
        except yaml.YAMLError as yea:
            raise ValueError(yea)
    else:
        if env is None:
            env = {}

    if not isinstance(env, dict):
        raise ValueError(_('The environment is not a valid '
                           'YAML mapping data type.'))
    return validate(env)


def validate(env):
    for param in env:
        if param not in SECTIONS:
            raise ValueError(_('environment has wrong section "%s"') % param)
        if env[param] is None:
            raise ValueError(_('environment has empty section "%s"') % param)

    return env


def default_for_missing(env):
    """Checks a parsed environment for missing sections."""
    for param in SECTIONS:
        if param not in env and param != PARAMETER_MERGE_STRATEGIES:
            if param in (ENCRYPTED_PARAM_NAMES, EVENT_SINKS):
                env[param] = []
            else:
                env[param] = {}



#    Copyright (c) 2018 AT&T Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine.resources.openstack.neutron import neutron
from heat.engine import support
from heat.engine import translation


class TapService(neutron.NeutronResource):
    """A resource for neutron tap-as-a-service tap-service.

    This plug-in requires neutron-taas. So to enable this
    plug-in, install this library and restart the heat-engine.

    A Tap-Service represents the port on which the mirrored traffic is
    delivered. Any VM that uses the mirrored data is attached to this port.
    """

    required_service_extension = 'taas'

    entity = 'tap_service'

    support_status = support.SupportStatus(version='12.0.0')

    PROPERTIES = (
        NAME, DESCRIPTION, PORT,
        ) = (
        'name', 'description', 'port',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name for the Tap-Service.'),
            default="",
            update_allowed=True
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description for the Tap-Service.'),
            default="",
            update_allowed=True
        ),
        PORT: properties.Schema(
            properties.Schema.STRING,
            _('ID or name of the tap-service neutron port.'),
            constraints=[constraints.CustomConstraint('neutron.port')],
            required=True,
        ),
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.PORT],
                client_plugin=self.client_plugin(),
                finder='find_resourceid_by_name_or_id',
                entity='port'
            )
        ]

    def _show_resource(self):
        return self.client_plugin().show_ext_resource('tap_service',
                                                      self.resource_id)

    def handle_create(self):
        props = self.prepare_properties(self.properties,
                                        self.physical_resource_name())
        props['port_id'] = props.pop(self.PORT)
        ts = self.client_plugin().create_ext_resource('tap_service',
                                                      props)
        self.resource_id_set(ts['id'])

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            self.prepare_update_properties(prop_diff)
            self.client_plugin().update_ext_resource('tap_service',
                                                     prop_diff,
                                                     self.resource_id)

    def handle_delete(self):
        if self.resource_id is None:
            return
        with self.client_plugin().ignore_not_found:
            self.client_plugin().delete_ext_resource('tap_service',
                                                     self.resource_id)

    def check_create_complete(self, data):
        return self.client_plugin().check_ext_resource_status(
            'tap_service', self.resource_id)

    def check_update_complete(self, prop_diff):
        if prop_diff:
            return self.client_plugin().check_ext_resource_status(
                'tap_service', self.resource_id)
        return True

    def check_delete_complete(self, data):
        if self.resource_id is None:
            return True

        with self.client_plugin().ignore_not_found:
            try:
                if self.client_plugin().check_ext_resource_status(
                        'tap_service', self.resource_id):
                    self.client_plugin().delete_ext_resource(
                        'tap_service', self.resource_id)
            except exception.ResourceInError:
                # Still try to delete tap resource in error state
                self.client_plugin().delete_ext_resource('tap_service',
                                                         self.resource_id)
            return False

        return True


def resource_mapping():
    return {
        'OS::Neutron::TaaS::TapService': TapService,
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstack.config import cloud_region
from openstack import connection
from openstack import exceptions
import os_service_types

from heat.common import config
from heat.engine.clients import client_plugin
from heat.engine.clients import os as os_client
from heat.engine import constraints
import heat.version

CLIENT_NAME = 'openstack'


class OpenStackSDKPlugin(os_client.ExtensionMixin,
                         client_plugin.ClientPlugin):

    exceptions_module = exceptions

    service_types = [NETWORK] = ['network']

    def _create(self, version=None):
        config = cloud_region.from_session(
            # TODO(mordred) The way from_session calculates a cloud name
            # doesn't interact well with the mocks in the test cases. The
            # name is used in logging to distinguish requests made to different
            # clouds. For now, set it to local - but maybe find a way to set
            # it to something more meaningful later.
            name='local',
            session=self.context.keystone_session,
            config=self._get_service_interfaces(),
            region_name=self._get_region_name(),
            app_name='heat',
            app_version=heat.version.version_info.version_string(),
            **self._get_additional_create_args(version))
        return connection.Connection(config=config)

    def _get_additional_create_args(self, version):
        return {}

    def _get_service_interfaces(self):
        interfaces = {}
        if not os_service_types:
            return interfaces
        types = os_service_types.ServiceTypes()
        for name, _ in config.list_opts():
            if not name or not name.startswith('clients_'):
                continue
            project_name = name.split("_", 1)[0]
            service_data = types.get_service_data_for_project(project_name)
            if not service_data:
                continue
            service_type = service_data['service_type']
            interfaces[service_type + '_interface'] = self._get_client_option(
                service_type, 'endpoint_type')
        return interfaces

    def is_not_found(self, ex):
        return isinstance(ex, exceptions.NotFoundException)

    def find_network_segment(self, value):
        return self.client().network.find_segment(value).id

    def find_network_port(self, value):
        return self.client().network.find_port(value).id

    def find_network_ip(self, value):
        return self.client().network.find_ip(value).id

    # TODO(tkajinam): This should be generalized when we onboard more services
    #                 requiring extension detection.
    @os_client.MEMOIZE_EXTENSIONS
    def _list_extensions(self):
        extensions = self.client().network.extensions()
        return set(extension.alias for extension in extensions)


class SegmentConstraint(constraints.BaseCustomConstraint):

    expected_exceptions = (exceptions.ResourceNotFound,
                           exceptions.DuplicateResource)

    def validate_with_client(self, client, value):
        sdk_plugin = client.client_plugin(CLIENT_NAME)
        sdk_plugin.find_network_segment(value)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support

LOG = logging.getLogger(__name__)


class MistralExternalResource(resource.Resource):
    """A plugin for managing user-defined resources via Mistral workflows.

    This resource allows users to manage resources that are not known to Heat.
    The user may specify a Mistral workflow to handle each resource action,
    such as CREATE, UPDATE, or DELETE.

    The workflows may return an output named 'resource_id', which will be
    treated as the physical ID of the resource by Heat.

    Once the resource is created, subsequent workflow runs will receive the
    output of the last workflow execution in the 'heat_extresource_data' key
    in the workflow environment (accessible as ``env().heat_extresource_data``
    in the workflow).

    The template author may specify a subset of inputs as causing replacement
    of the resource when they change, as an alternative to running the
    UPDATE workflow.
    """

    support_status = support.SupportStatus(version='9.0.0')

    default_client_name = 'mistral'

    entity = 'executions'

    _ACTION_PROPERTIES = (
        WORKFLOW, PARAMS
    ) = (
        'workflow', 'params'
    )

    PROPERTIES = (
        EX_ACTIONS,
        INPUT,
        DESCRIPTION,
        REPLACE_ON_CHANGE,
        ALWAYS_UPDATE
    ) = (
        'actions',
        'input',
        'description',
        'replace_on_change_inputs',
        'always_update'
    )

    ATTRIBUTES = (
        OUTPUT,
    ) = (
        'output',
    )

    _action_properties_schema = properties.Schema(
        properties.Schema.MAP,
        _('Dictionary which defines the workflow to run and its params.'),
        schema={
            WORKFLOW: properties.Schema(
                properties.Schema.STRING,
                _('Workflow to execute.'),
                required=True,
                constraints=[
                    constraints.CustomConstraint('mistral.workflow')
                ],
            ),
            PARAMS: properties.Schema(
                properties.Schema.MAP,
                _('Workflow additional parameters. If workflow is reverse '
                  'typed, params requires "task_name", which defines '
                  'initial task.'),
                default={}
            ),
        }
    )

    properties_schema = {
        EX_ACTIONS: properties.Schema(
            properties.Schema.MAP,
            _('Resource action which triggers a workflow execution.'),
            schema={
                resource.Resource.CREATE: _action_properties_schema,
                resource.Resource.UPDATE: _action_properties_schema,
                resource.Resource.SUSPEND: _action_properties_schema,
                resource.Resource.RESUME: _action_properties_schema,
                resource.Resource.DELETE: _action_properties_schema,
            },
            required=True
        ),
        INPUT: properties.Schema(
            properties.Schema.MAP,
            _('Dictionary which contains input for the workflows.'),
            update_allowed=True,
            default={}
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Workflow execution description.'),
            default='Heat managed'
        ),
        REPLACE_ON_CHANGE: properties.Schema(
            properties.Schema.LIST,
            _('A list of inputs that should cause the resource to be replaced '
              'when their values change.'),
            default=[]
        ),
        ALWAYS_UPDATE: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Triggers UPDATE action execution even if input is '
              'unchanged.'),
            default=False
        ),
    }

    attributes_schema = {
        OUTPUT: attributes.Schema(
            _('Output from the execution.'),
            type=attributes.Schema.MAP
        ),
    }

    def _check_execution(self, action, execution_id):
        """Check execution status.

        Returns False if in IDLE, RUNNING or PAUSED
        returns True if in SUCCESS
        raises ResourceFailure if in ERROR, CANCELLED
        raises ResourceUnknownState otherwise.
        """
        execution = self.client().executions.get(execution_id)
        LOG.debug('Mistral execution %(id)s is in state '
                  '%(state)s' % {'id': execution_id,
                                 'state': execution.state})

        if execution.state in ('IDLE', 'RUNNING', 'PAUSED'):
            return False, {}

        if execution.state in ('SUCCESS',):
            return True, jsonutils.loads(execution.output)

        if execution.state in ('ERROR', 'CANCELLED'):
            raise exception.ResourceFailure(
                exception_or_error=execution.state_info,
                resource=self,
                action=action)

        raise exception.ResourceUnknownStatus(
            resource_status=execution.state,
            result=_('Mistral execution is in unknown state.'))

    def _handle_action(self, action, inputs=None):
        action_data = self.properties[self.EX_ACTIONS].get(action)
        if action_data:
            # bring forward output from previous executions into env
            if self.resource_id:
                old_outputs = jsonutils.loads(self.data().get('outputs', '{}'))
                action_env = action_data[self.PARAMS].get('env', {})
                action_env['heat_extresource_data'] = old_outputs
                action_data[self.PARAMS]['env'] = action_env
            # inputs is not None when inputs changed on stack UPDATE
            if not inputs:
                inputs = self.properties[self.INPUT]
            execution = self.client().executions.create(
                action_data[self.WORKFLOW],
                workflow_input=jsonutils.dumps(inputs),
                description=self.properties[self.DESCRIPTION],
                **action_data[self.PARAMS])
            LOG.debug('Mistral execution %(id)s params set to '
                      '%(params)s' % {'id': execution.id,
                                      'params': action_data[self.PARAMS]})
            return execution.id

    def _check_action(self, action, execution_id):
        success = True
        # execution_id is None when no data is available for a given action
        if execution_id:
            rsrc_id = execution_id
            success, output = self._check_execution(action, execution_id)
            # merge output with outputs of previous executions
            outputs = jsonutils.loads(self.data().get('outputs', '{}'))
            outputs.update(output)
            self.data_set('outputs', jsonutils.dumps(outputs))
            # set resource id using output, if found
            if output.get('resource_id'):
                rsrc_id = output.get('resource_id')
                LOG.debug('ExternalResource id set to %(rid)s from Mistral '
                          'execution %(eid)s output' % {'eid': execution_id,
                                                        'rid': rsrc_id})
            self.resource_id_set(str(rsrc_id)[:255])
        return success

    def _resolve_attribute(self, name):
        if self.resource_id and name == self.OUTPUT:
            return self.data().get('outputs')

    def _needs_update(self, after, before, after_props, before_props,
                      prev_resource, check_init_complete=True):
        # check if we need to force replace first
        old_inputs = before_props[self.INPUT]
        new_inputs = after_props[self.INPUT]
        for i in after_props[self.REPLACE_ON_CHANGE]:
            if old_inputs.get(i) != new_inputs.get(i):
                LOG.debug('Replacing ExternalResource %(id)s instead of '
                          'updating due to change to input "%(i)s"' %
                          {"id": self.resource_id,
                           "i": i})
                raise resource.UpdateReplace(self)
        # honor always_update if found
        if self.properties[self.ALWAYS_UPDATE]:
            return True
        # call super in all other scenarios
        else:
            return super(MistralExternalResource,
                         self)._needs_update(after,
                                             before,
                                             after_props,
                                             before_props,
                                             prev_resource,
                                             check_init_complete)

    def handle_create(self):
        return self._handle_action(self.CREATE)

    def check_create_complete(self, execution_id):
        return self._check_action(self.CREATE, execution_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        new_inputs = prop_diff.get(self.INPUT)
        return self._handle_action(self.UPDATE, new_inputs)

    def check_update_complete(self, execution_id):
        return self._check_action(self.UPDATE, execution_id)

    def handle_suspend(self):
        return self._handle_action(self.SUSPEND)

    def check_suspend_complete(self, execution_id):
        return self._check_action(self.SUSPEND, execution_id)

    def handle_resume(self):
        return self._handle_action(self.RESUME)

    def check_resume_complete(self, execution_id):
        return self._check_action(self.RESUME, execution_id)

    def handle_delete(self):
        return self._handle_action(self.DELETE)

    def check_delete_complete(self, execution_id):
        return self._check_action(self.DELETE, execution_id)


def resource_mapping():
    return {
        'OS::Mistral::ExternalResource': MistralExternalResource
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

from heat.common.i18n import _

from heat.common import exception
from heat.engine import function


_in_progress = object()


class Conditions(object):
    def __init__(self, conditions_dict):
        assert isinstance(conditions_dict, collections.abc.Mapping)
        self._conditions = conditions_dict
        self._resolved = {}

    def validate(self):
        for name, cond in self._conditions.items():
            self._check_condition_type(name, cond)
            function.validate(cond)

    def _resolve(self, condition_name):
        resolved = function.resolve(self._conditions[condition_name])
        self._check_condition_type(condition_name, resolved)
        return resolved

    def _check_condition_type(self, condition_name, condition_defn):
        if not isinstance(condition_defn, (bool, function.Function)):
            msg_data = {'cd': condition_name, 'definition': condition_defn}
            message = _('The definition of condition "%(cd)s" is invalid: '
                        '%(definition)s') % msg_data
            raise exception.StackValidationFailed(
                error='Condition validation error',
                message=message)

    def is_enabled(self, condition_name):
        if condition_name is None:
            return True

        if isinstance(condition_name, bool):
            return condition_name

        if not (isinstance(condition_name, str) and
                condition_name in self._conditions):
            raise ValueError(_('Invalid condition "%s"') % condition_name)

        if condition_name not in self._resolved:
            self._resolved[condition_name] = _in_progress
            self._resolved[condition_name] = self._resolve(condition_name)

        result = self._resolved[condition_name]

        if result is _in_progress:
            message = _('Circular definition for condition '
                        '"%s"') % condition_name
            raise exception.StackValidationFailed(
                error='Condition validation error',
                message=message)

        return result

    def __repr__(self):
        return 'Conditions(%r)' % self._conditions



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import exception
from heat.common import template_format
from heat.engine.clients.os import mistral as mistral_client
from heat.engine.resources.openstack.vitrage.vitrage_template import \
    VitrageTemplate
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


vitrage_template = '''
heat_template_version: rocky
resources:
     execute_healing:
       type: OS::Vitrage::Template
       description: Execute Mistral healing workflow if instance is down
       properties:
         template_file: execute_healing_on_instance_down.yaml
         template_params:
           instance_alarm_name: Instance down
           instance_id: 1233e48c-62ee-470e-8d4a-adff30211b5d
           workflow_name: autoheal
           heat_stack_id: 12cc6d3e-f801-4422-b2a0-43cedacb4eb5

'''


class TestVitrageTemplate(common.HeatTestCase):
    def setUp(self):
        super(TestVitrageTemplate, self).setUp()
        self.ctx = utils.dummy_context()
        template = template_format.parse(vitrage_template)
        self.stack = utils.parse_stack(template, stack_name='test_stack')

        resource_defs = self.stack.t.resource_definitions(self.stack)
        self.resource_def = resource_defs['execute_healing']

        self.vitrage = mock.Mock()
        self.patchobject(VitrageTemplate, 'client', return_value=self.vitrage)

        self.patches = []
        self.patches.append(mock.patch.object(mistral_client,
                                              'mistral_base'))
        self.patches.append(mock.patch.object(
            mistral_client.MistralClientPlugin, '_create'))
        for patch in self.patches:
            patch.start()

        self.mistral_client = \
            mistral_client.MistralClientPlugin(context=self.ctx)

    def tearDown(self):
        super(TestVitrageTemplate, self).tearDown()
        for patch in self.patches:
            patch.stop()

    def test_create(self):
        template = self._create_resource(
            'execute_healing', self.resource_def, self.stack)
        expected_state = (template.CREATE, template.COMPLETE)

        # Verify the creation succeeded
        self.assertEqual(expected_state, template.state)
        self.assertEqual('2fddb683-e32c-4a9b-b8c8-df59af1f5a1a',
                         template.get_reference_id())

    def test_validate(self):
        self.vitrage.template.validate.return_value = {
            "results": [
                {
                    "status": "validation OK",
                    "file path": "/tmp/tmpNUEgE3",
                    "message": "Template validation is OK",
                    "status code": 0,
                    "description": "Template validation"
                }
            ]
        }

        # No result for a valid template
        template = \
            VitrageTemplate('execute_healing', self.resource_def, self.stack)
        scheduler.TaskRunner(template.validate)()
        self.vitrage.template.validate.assert_called_once()

    def test_validate_vitrage_validate_wrong_format(self):
        """wrong result format for vitrage templete validate"""
        template = \
            VitrageTemplate('execute_healing', self.resource_def, self.stack)

        # empty return value
        self.vitrage.template.validate.return_value = {}
        self.assertRaises(exception.StackValidationFailed,
                          scheduler.TaskRunner(template.validate))

        # empty 'results'
        self.vitrage.template.validate.return_value = {
            "results": []
        }
        self.assertRaises(exception.StackValidationFailed,
                          scheduler.TaskRunner(template.validate))

        # too many 'results'
        self.vitrage.template.validate.return_value = {
            "results": [
                {
                    "status": "validation OK",
                    "file path": "/tmp/tmpNUEgE3",
                    "message": "Template validation is OK",
                    "status code": 0,
                    "description": "Template validation"
                },
                {
                    "status": "validation OK",
                    "file path": "/tmp/tmpNUEgE3",
                    "message": "Template validation is OK",
                    "status code": 0,
                    "description": "Template validation"
                },
            ]
        }
        self.assertRaises(exception.StackValidationFailed,
                          scheduler.TaskRunner(template.validate))

        # no 'status code'
        self.vitrage.template.validate.return_value = {
            "results": [
                {
                    "status": "validation OK",
                    "file path": "/tmp/tmpNUEgE3",
                    "message": "Template validation is OK",
                    "description": "Template validation"
                }
            ]
        }
        self.assertRaises(exception.StackValidationFailed,
                          scheduler.TaskRunner(template.validate))

    def test_validate_vitrage_validation_failed(self):
        template = \
            VitrageTemplate('execute_healing', self.resource_def, self.stack)

        self.vitrage.template.validate.return_value = {
            "results": [
                {
                    "status": "validation failed",
                    "file path": "/tmp/tmpNUEgE3",
                    "status code": 163,
                    "message": "Failed to resolve parameter",
                    "description": "Template content validation"
                }
            ]
        }
        self.assertRaises(exception.StackValidationFailed,
                          scheduler.TaskRunner(template.validate))

    def _create_resource(self, name, snippet, stack):
        template = VitrageTemplate(name, snippet, stack)
        self.vitrage.template.add.return_value = [{
            'status': 'LOADING',
            'uuid': '2fddb683-e32c-4a9b-b8c8-df59af1f5a1a',
            'status details': 'Template validation is OK',
            'date': '2019-02-20 16:36:17.240976',
            'type': 'standard',
            'name': 'Stack40-execute_healing-4ri7d3vlwp5w'
        }]
        scheduler.TaskRunner(template.create)()
        return template



# Copyright 2014 Intel Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""RawTemplate object."""

import copy

from oslo_config import cfg
from oslo_log import log as logging
from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from heat.common import crypt
from heat.common import environment_format as env_fmt
from heat.db import api as db_api
from heat.objects import base as heat_base
from heat.objects import fields as heat_fields

LOG = logging.getLogger(__name__)


@heat_base.HeatObjectRegistry.register
class RawTemplate(
    heat_base.HeatObject,
    base.VersionedObjectDictCompat,
    base.ComparableVersionedObject,
):
    # Version 1.0: Initial version
    # Version 1.1: Added files_id
    VERSION = '1.1'

    fields = {
        'id': fields.IntegerField(),
        # TODO(cwolfe): remove deprecated files in future release
        'files': heat_fields.JsonField(nullable=True),
        'files_id': fields.IntegerField(nullable=True),
        'template': heat_fields.JsonField(),
        'environment': heat_fields.JsonField(),
    }

    @staticmethod
    def from_db_object(context, tpl, db_tpl):
        for field in tpl.fields:
            tpl[field] = db_tpl[field]

        tpl.environment = copy.deepcopy(tpl.environment)
        # If any of the parameters were encrypted, then decrypt them
        if (tpl.environment is not None and
                env_fmt.ENCRYPTED_PARAM_NAMES in tpl.environment):
            parameters = tpl.environment[env_fmt.PARAMETERS]
            encrypted_param_names = tpl.environment[
                env_fmt.ENCRYPTED_PARAM_NAMES]

            for param_name in encrypted_param_names:
                if (isinstance(parameters[param_name], (list, tuple)) and
                        len(parameters[param_name]) == 2):
                    method, enc_value = parameters[param_name]
                    value = crypt.decrypt(method, enc_value)
                else:
                    value = parameters[param_name]
                    LOG.warning(
                        'Encountered already-decrypted data while attempting '
                        'to decrypt parameter %s. Please file a Heat bug so '
                        'this can be fixed.', param_name)
                parameters[param_name] = value
            tpl.environment[env_fmt.PARAMETERS] = parameters

        tpl._context = context
        tpl.obj_reset_changes()
        return tpl

    @classmethod
    def get_by_id(cls, context, template_id):
        raw_template_db = db_api.raw_template_get(context, template_id)
        return cls.from_db_object(context, cls(), raw_template_db)

    @classmethod
    def encrypt_hidden_parameters(cls, tmpl):
        if cfg.CONF.encrypt_parameters_and_properties:
            for param_name in tmpl.env.params.keys():
                if not tmpl.param_schemata()[param_name].hidden:
                    continue
                clear_text_val = tmpl.env.params.get(param_name)
                tmpl.env.params[param_name] = crypt.encrypt(clear_text_val)
                if param_name not in tmpl.env.encrypted_param_names:
                    tmpl.env.encrypted_param_names.append(param_name)

    @classmethod
    def create(cls, context, values):
        return cls.from_db_object(context, cls(),
                                  db_api.raw_template_create(context, values))

    @classmethod
    def update_by_id(cls, context, template_id, values):
        # Only save template files in the new raw_template_files
        # table, not in the old location of raw_template.files
        if 'files_id' in values and values['files_id']:
            values['files'] = None
        return cls.from_db_object(
            context, cls(),
            db_api.raw_template_update(context, template_id, values))

    @classmethod
    def delete(cls, context, template_id):
        db_api.raw_template_delete(context, template_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from heat.engine import translation


class AvailabilityZone(resource.Resource):
    """A resource for creating octavia Availability Zones.

    This resource creates and manages octavia Availability Zones,
    which allows to tune Load Balancers' capabilities.
    """

    default_client_name = 'octavia'

    support_status = support.SupportStatus(version='24.0.0')

    PROPERTIES = (
        DESCRIPTION, ENABLED, NAME, AVAILABILITY_ZONE_PROFILE
    ) = (
        'description', 'enabled', 'name', 'availability_zone_profile'
    )

    ATTRIBUTES = (
        AVAILABILITY_ZONE_PROFILE_ID_ATTR,
    ) = (
        'availability_zone_profile_id',
    )

    properties_schema = {
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description of this Availability Zone.'),
            update_allowed=True,
            default=''
        ),
        ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('If the resource if available for use.'),
            update_allowed=True,
            default=True,
        ),
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of this Availability Zone.'),
            update_allowed=True
        ),
        AVAILABILITY_ZONE_PROFILE: properties.Schema(
            properties.Schema.STRING,
            _('The ID or the name of the Availability Zone Profile.'),
            required=True,
            constraints=[
                constraints.CustomConstraint('octavia.availabilityzoneprofile')
            ]
        ),
    }

    attributes_schema = {
        AVAILABILITY_ZONE_PROFILE_ID_ATTR: attributes.Schema(
            _('The ID of the availability zone profile.'),
            type=attributes.Schema.STRING,
        )
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.AVAILABILITY_ZONE_PROFILE],
                client_plugin=self.client_plugin(),
                finder='get_availabilityzoneprofile'
            )
        ]

    def _prepare_args(self, properties):
        props = dict((k, v) for k, v in properties.items()
                     if v is not None)
        if self.NAME not in props:
            props[self.NAME] = self.physical_resource_name()
        props['availability_zone_profile_id'] = props.pop(
            self.AVAILABILITY_ZONE_PROFILE
        )
        return props

    def handle_create(self):
        props = self._prepare_args(self.properties)

        availability_zone = self.client().availabilityzone_create(
            json={'availability_zone': props})['availability_zone']
        self.resource_id_set(availability_zone.get('name'))

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            if self.NAME in prop_diff and prop_diff[self.NAME] is None:
                prop_diff[self.NAME] = self.physical_resource_name()
            self.client().availabilityzone_set(
                self.resource_id,
                json={'availability_zone': prop_diff})

    def handle_delete(self):
        with self.client_plugin().ignore_not_found:
            self.client().availabilityzone_delete(self.resource_id)
            return True

    def _show_resource(self):
        return self.client().availabilityzone_show(self.resource_id)


def resource_mapping():
    return {
        'OS::Octavia::AvailabilityZone': AvailabilityZone
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from designateclient import client as designate_client

from heat.common import exception as heat_exception
from heat.engine.clients.os import designate as client
from heat.tests import common


class DesignateClientPluginTest(common.HeatTestCase):

    @mock.patch.object(designate_client, 'Client')
    def test_client(self, client_designate):
        context = mock.Mock()
        session = mock.Mock()
        context.keystone_session = session
        client_plugin = client.DesignateClientPlugin(context)
        self.patchobject(client_plugin, '_get_region_name',
                         return_value='region1')
        client_plugin.client()

        # Make sure proper client is created with expected args
        client_designate.assert_called_once_with(
            endpoint_type='publicURL', service_type='dns',
            session=session, region_name='region1',
            version='2')


class DesignateZoneConstraintTest(common.HeatTestCase):

    def test_expected_exceptions(self):
        self.assertEqual((heat_exception.EntityNotFound,),
                         client.DesignateZoneConstraint.expected_exceptions,
                         "DesignateZoneConstraint expected exceptions error")

    def test_constrain(self):
        constrain = client.DesignateZoneConstraint()
        client_mock = mock.MagicMock()
        client_plugin_mock = mock.MagicMock()
        client_plugin_mock.get_zone_id.return_value = None
        client_mock.client_plugin.return_value = client_plugin_mock

        self.assertIsNone(constrain.validate_with_client(client_mock,
                                                         'zone_1'))

        client_plugin_mock.get_zone_id.assert_called_once_with('zone_1')
